@@ -1,32 +1,20 @@
 /******************************************************************************
-** $Id: config.h,v 2.20 1996/02/22 21:43:51 gerd Exp gerd $
+** $Id: config.h,v 1.1 1997/12/06 22:56:00 gerd Exp gerd $
 **=============================================================================
 ** 
 ** This file is part of BibTool.
 ** It is distributed under the GNU General Public License.
 ** See the file COPYING for details.
 ** 
-** (c) 1996 Gerd Neugebauer
+** (c) 1996-1997 Gerd Neugebauer
 ** 
 ** Net: gerd@informatik.uni-koblenz.de
 ** 
 ******************************************************************************/
 
-/*-----------------------------------------------------------------------------
-**  Define the following macro if this is compiled on a MS-DOS-like system.
-**  (MSDOS, OS/2, Windows, Atari)
-**  This is also done in the makefile. Do not define it here if you use the
-**  makefile.
-**  It affects the way file names are handled.
-*/
-/*#define MSDOS /**/
-/*-----------------------------------------------------------------------------
-**  Define the following macro if this is compiled on an Amiga.
-**  This is also done in the makefile. Do not define it here if you use the
-**  makefile.
-**  It affects the way file names are handled.
-*/
-/*#define AMIGA /**/
+/******************************************************************************
+**			    Configuration Section
+******************************************************************************/
 
 /*-----------------------------------------------------------------------------
 **  Define the default value of the resource search path.
@@ -45,8 +33,8 @@
 **  If the following macro is defined then the standard @types are provided
 **  automatically.
 **  The value is a comma separated list of strings.
-**  Note that COMMENT, PREAMBLE, and STRING are hard wired and need not to be
-**  in this list!
+**  Note that COMMENT, PREAMBLE, STRING, ALIAS, MODIFY, and INCLUDE
+**  are hard wired and need not to be in this list!
 */
 #define INITIALIZE_BIBTEX_ENTRIES					 \
 	"Article"	,"Book"		,"Booklet"	,"Conference"	,\
@@ -167,16 +155,17 @@
 
 /*-----------------------------------------------------------------------------
 **  Maximal number of user defined formats.
-**  At least 2 are reqiured by BibTool!
+**  At least 2 are required by BibTool!
 */
 #define NUMBER_OF_FORMATS 128
 
 /*-----------------------------------------------------------------------------
 **  Size of the hash table.
 **  This should be a prime number. Don't use a value which is too small.
-**  307 is just ok at the moment.
+**  307 is just ok if you consider databases with up to 100 entries.
+**  1021 is for larger databases.
 */
-#define HASH_TABLE_SIZE 307
+#define HASH_TABLE_SIZE 1021
 
 /*-----------------------------------------------------------------------------
 **  Define the next macro to enable output of the symbol table.
@@ -184,51 +173,6 @@
 */
 /* #define SYMBOL_DUMP /**/
 
-#ifndef __STDC__
-/*-----------------------------------------------------------------------------
-**  The following section contains (rudimentary) support for non-ANSI
-**  C compilers. If you have an ANSI C compiler simply ignore them.
-**
-**-----------------------------------------------------------------------------
-**  Don't change this! unless you have a non-ANSI C compiler which
-**  knows the const keyword (very unlikely).
-**  Otherwise just comment it away.
-*/
-#define const
-
-/*-----------------------------------------------------------------------------
-**  Define the following macro if you are using a non-ANSI C compiler
-**  which has stdlib.h
-*/
-#define HAS_stdlib
-
-/*-----------------------------------------------------------------------------
-**  Define the following macro if you are using a non-ANSI C compiler
-**  which has a getenv() library function.
-**  Try it with this macro defined and undefine it if the linker complains.
-*/
-#define HAS_getenv
-
-/*-----------------------------------------------------------------------------
-**  Define here the type of the argument of malloc if not declared otherwise.
-**  This value is used in non-ANSI compilers without stdlib.h
-*/
-#define SIZE_T unsigned int
-
-/*-----------------------------------------------------------------------------
-**  Define here the type of malloc() and realloc if not declared otherwise.
-**  Reasonable values are char* and void*
-**  This value is used in non-ANSI compilers without stdlib.h
-*/
-#define VoidPTR char*
-
-/*-----------------------------------------------------------------------------
-**  Define the following macro to use strings.h instead of string.h and
-**  index() instead of strchr()
-*/
-#define OLD_STYLE_STRINGS
-
-#endif
 
 #ifdef MSDOS
 /*-----------------------------------------------------------------------------
@@ -247,7 +191,7 @@
 */
 #define EMTEXDIR		"EMTEXDIR"
 /*-----------------------------------------------------------------------------
-**  The following macro holds the default in no environment variable is found.
+**  The following macro holds the default if no environment variable is found.
 */
 #define EMTEXTDIR_DEFAULT	"\\emtex"
 /*-----------------------------------------------------------------------------
