@@ -1,63 +1,46 @@
 /******************************************************************************
-** $Id: macros.c,v 2.18 1996/02/22 21:43:51 gerd Exp gerd $
+** $Id: macros.c,v 2.26 1997/12/06 22:56:00 gerd Exp gerd $
 **=============================================================================
 ** 
 ** This file is part of BibTool.
 ** It is distributed under the GNU General Public License.
 ** See the file COPYING for details.
 ** 
-** (c) 1996 Gerd Neugebauer
+** (c) 1996-1997 Gerd Neugebauer
 ** 
 ** Net: gerd@informatik.uni-koblenz.de
 ** 
 ******************************************************************************/
 
-#include "bibtool.h"
-#include "symbols.h"
-#include "error.h"
-#include "type.h"
-#include "macros.h"
-#include "sbuffer.h"
+#include <bibtool/general.h>
+#include <bibtool/symbols.h>
+#include <bibtool/error.h>
+#include <bibtool/macros.h>
+#include <bibtool/sbuffer.h>
+#include "config.h"
 
 /*****************************************************************************/
 /* Internal Programs                                                         */
 /*===========================================================================*/
 
-#ifdef __STDC__
-#define _ARG(A) A
-#else
-#define _ARG(A) ()
-#endif
- Macro new_macro _ARG((char *name,char *val,Macro next,int count));/* macros.c*/
- char * get_item _ARG((char * name,int type));	   /* macros.c               */
- char * get_key_name _ARG((char *symbol));	   /* macros.c               */
- char * look_macro _ARG((char *name,int add));	   /* macros.c               */
- static char * get_mapped_or_cased _ARG((char * name,Macro mac,int type));/* macros.c*/
- void def_field_type _ARG((char * s));		   /* macros.c               */
- void def_item _ARG((char * name,char * value));   /* macros.c               */
- void def_macro _ARG((char *name,char *val,int count));/* macros.c           */
- void dump_mac _ARG((char *fname,int allp));	   /* macros.c               */
- void init_macros _ARG((void));			   /* macros.c               */
- void save_key _ARG((char * symbol,char * key));   /* macros.c               */
-
 /*****************************************************************************/
 /* External Programs                                                         */
 /*===========================================================================*/
 
 /*---------------------------------------------------------------------------*/
 
- static Macro macros = MacNull;			   /*                        */
+ static Macro macros = MacroNULL;		   /*                        */
 
 /*-----------------------------------------------------------------------------
 ** Function:	new_macro()
 ** Purpose:	Allocate a new macro structure and fill it with initial values.
-**		Upon failure exit() is called.
+**		Upon failure an error is raised and |exit()| is called.
 ** Arguments:
-**	name
-**	val
-**	next
-**	count
-** Returns:	The new Macro
+**	name	Name of the macro. This must be a symbol.
+**	val	The value of the macro. This must be a symbol.
+**	next	The next pointer of the |Macro| structure.
+**	count	The initial reference count.
+** Returns:	The new |Macro|.
 **___________________________________________________			     */
 Macro new_macro(name,val,next,count)		   /*                        */
   register char  *name;				   /*                        */
@@ -66,56 +49,91 @@ Macro new_macro(name,val,next,count)		   /*                        */
   Macro          next;				   /*                        */
 { register Macro new;				   /*                        */
  						   /*                        */
-  if ( (new=(Macro)malloc(sizeof(SMacro))) == MacNull )/*                    */
+  if ( (new=(Macro)malloc(sizeof(SMacro))) == MacroNULL )/*                  */
   { OUT_OF_MEMORY("Macro"); }      		   /*                        */
  						   /*                        */
   MacroName(new)  = name;			   /*                        */
-  MacroVal(new)   = val ;			   /*                        */
+  MacroValue(new) = val ;			   /*                        */
   MacroCount(new) = count;			   /*                        */
   NextMacro(new)  = next;			   /*                        */
   return(new);					   /*                        */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:	def_macro()
-** Purpose:	Define a macro. Give a warning if it is already defined.
+** Function:	free_macro()
+** Purpose:	Free a list of macros. The memory allocated for the
+**		|Macro| given as argument and all struictures
+**		reachable via the |NextMacro| pointer are released.
 ** Arguments:
-**	name
-**	val
-**	count
+**	mac	First Macro to release.
 ** Returns:	nothing
 **___________________________________________________			     */
-void def_macro(name,val,count)			   /*                        */
+void free_macro(mac)				   /*                        */
+  Macro mac;					   /*                        */
+{ Macro next;					   /*                        */
+ 						   /*                        */
+  while (mac)					   /*                        */
+  { next = NextMacro(mac);			   /*                        */
+    free(mac);					   /*                        */
+    mac = next;					   /*                        */
+  }						   /*                        */
+}						   /*------------------------*/
+
+/*-----------------------------------------------------------------------------
+** Function:	def_macro()
+** Purpose:	Define or undefine a macro.
+** Arguments:
+**	name	name of the macro.
+**	val	NULL or the value of the new macro
+**	count	initial count for the macro.
+** Returns:	nothing
+**___________________________________________________			     */
+int def_macro(name,val,count)			   /*                        */
   char           *name;				   /*                        */
   char           *val;				   /*                        */
   int		 count;				   /*                        */
 { register Macro *mp;				   /*                        */
  						   /*                        */
   name = symbol(name);				   /*                        */
-  val  = symbol(val);				   /*                        */
+  if ( val ) val  = symbol(val);		   /*                        */
  						   /*                        */
   for ( mp   = &macros;				   /*                        */
-        *mp != MacNull;				   /*                        */
+        *mp != MacroNULL;			   /*                        */
 	mp   = &NextMacro(*mp)			   /*                        */
       )						   /*                        */
   { if ( MacroName(*mp) == name )		   /*                        */
-    { WARNING3("Macro '",name,"' is already defined.");/*                    */
-      MacroVal(*mp) = val;			   /*                        */
-      return;					   /*                        */
+    {						   /*                        */
+      if ( val )				   /*                        */
+      { MacroValue(*mp) = val; }		   /*                        */
+      else
+      { Macro mac = *mp;
+        *mp = NextMacro(*mp);
+	free(mac);
+      }
+      return 1;					   /*                        */
     }						   /*                        */
   }						   /*                        */
  						   /*                        */
-  macros = new_macro(name,val,macros,count);	   /*                        */
+  if ( val )					   /*                        */
+  { macros = new_macro(name,val,macros,count); }   /*                        */
+  return 0;					   /*                        */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
 ** Function:	look_macro()
-** Purpose:	Return the value of a macro.
-**		If the macro is undefined its name is returned.
+** Purpose:	Return the value of a macro if it is defined. This
+**		value is a symbol. 
+**		If the macro is undefined then |NULL| is returned.  In
+**		this case the value of |add| determines whether or not
+**		the macro shpould be defined. If it is less than 0
+**		then no new macros is defined. Otherwise a new macro
+**		is defined. The value is the empty string and the
+**		initial reference count is |add|.
 ** Arguments:
-**	name
-**	add
-** Returns:	The value.
+**	name	The name of the macros to find. This needs not to be a symbol.
+**	add	Initial reference count or indicator that no new macro
+**		is required.
+** Returns:	The value or |NULL|.
 **___________________________________________________			     */
 char * look_macro(name,add)			   /*                        */
   char           *name;				   /*                        */
@@ -125,26 +143,59 @@ char * look_macro(name,add)			   /*                        */
   name = symbol(name);				   /*                        */
 						   /*                        */
   for ( mp   = &macros;				   /*                        */
-        *mp != MacNull;				   /*                        */
+        *mp != MacroNULL;			   /*                        */
 	mp   = &NextMacro(*mp)			   /*                        */
       )						   /*                        */
   { if ( MacroName(*mp) == name )		   /*                        */
     { if ( MacroCount(*mp) >= 0 )		   /*                        */
 	MacroCount(*mp) += add;			   /*                        */
-      return(MacroVal(*mp));			   /*                        */
+      return(MacroValue(*mp));			   /*                        */
     }						   /*                        */
   }						   /*                        */
-  def_macro(name,name,add);			   /*                        */
+  if ( add < 0 ) return NULL;			   /*                        */
+  def_macro(name,sym_empty,add);		   /*                        */
   WARNING3("Macro '",name,"' is undefined.");	   /*                        */
-  return(name);					   /*                        */
+  return NULL;					   /*                        */
+}						   /*------------------------*/
+
+/*-----------------------------------------------------------------------------
+** Function:	foreach_macro()
+** Purpose:	Apply a function to each macro in turn. The function
+**		is called with the name and the value of the macro. If
+**		it returns |FALSE| then the processing of further
+**		macros is suppressed.
+**
+**		The function given as argument is called with two
+**		string arguments. The first is the name of the macro
+**		and the second is its value. Both are symbols and must
+**		not be modified in any way.
+**
+**		The order of the enumeration of the macros is
+**		determined by the implementation. No specific
+**		assumptions should be made about this order.
+** Arguments:
+**	fct	Function to apply to each macro.
+** Returns:	nothing
+**___________________________________________________			     */
+void foreach_macro(fct)				   /*                        */
+  int (*fct) _ARG((char *,char *));		   /*                        */
+{ Macro mac;					   /*                        */
+  for ( mac=macros; 				   /*                        */
+	mac != MacroNULL; 			   /*                        */
+	mac = NextMacro(mac) )			   /*                        */
+  {						   /*                        */
+    if ( MacroValue(mac) &&			   /*                        */
+	 ! (*fct)(MacroName(mac),MacroValue(mac)) )/*                        */
+      return;					   /*                        */
+  }						   /*                        */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
 ** Function:	dump_mac()
 ** Purpose:	Write macros to a file.
 ** Arguments:
-**	fname	file name of the target file.
-**	allp	if ==0 only the used macros are written.
+**	fname	File name of the target file.
+**	allp	if == 0 only the used macros are written.
 ** Returns:	nothing
 **___________________________________________________			     */
 void dump_mac(fname,allp)			   /*                        */
@@ -161,24 +212,24 @@ void dump_mac(fname,allp)			   /*                        */
   }					   	   /*                        */
  						   /*                        */
   for ( mac=macros; 				   /*                        */
-	mac != MacNull; 			   /*                        */
+	mac != MacroNULL; 			   /*                        */
 	mac = NextMacro(mac) )			   /*                        */
   { if ( MacroCount(mac) > 0  ||		   /*                        */
          ( MacroCount(mac) >= 0 && allp) )	   /*                        */
-    { if ( MacroName(mac) == MacroVal(mac) )
+    { if ( MacroName(mac) == MacroValue(mac) )	   /*                        */
       { (void)fprintf(file,			   /*                        */
 		      "_string{ %s = %s } used: %d\n",/*                     */
 		      MacroName(mac),		   /*                        */
-		      MacroVal(mac),		   /*                        */
+		      MacroValue(mac),		   /*                        */
 		      MacroCount(mac));		   /*                        */
-      }
-      else
+      }						   /*                        */
+      else					   /*                        */
       { (void)fprintf(file,			   /*                        */
 		      "@STRING{ %s = %s } used: %d\n",/*                     */
 		      MacroName(mac),		   /*                        */
-		      MacroVal(mac),		   /*                        */
+		      MacroValue(mac),		   /*                        */
 		      MacroCount(mac));		   /*                        */
-      }
+      }						   /*                        */
     }						   /*                        */
   }						   /*                        */
  						   /*                        */
@@ -189,7 +240,14 @@ void dump_mac(fname,allp)			   /*                        */
 
 /*-----------------------------------------------------------------------------
 ** Function:	init_macros()
-** Purpose:	Initialize some macros from a table
+** Purpose:	Initialize some macros from a table defined in the
+**		configuration file or given as define to the C
+**		compiler. This function has to be caled to initialize
+**		the global macros.
+**
+**		Note that this function is for internal purposes
+**		only. The normal user should call |init_bibtool()|
+**		instead.
 ** Arguments:	none
 ** Returns:	nothing
 **___________________________________________________			     */
@@ -211,7 +269,7 @@ void init_macros()				   /*                        */
 /***									   ***/
 /*****************************************************************************/
 
- static Macro items = MacNull;		
+ static Macro items = MacroNULL;		
 
 /*-----------------------------------------------------------------------------
 ** Function:	def_item()
@@ -222,7 +280,7 @@ void init_macros()				   /*                        */
 **	value
 ** Returns:	nothing
 **___________________________________________________			     */
-void def_item(name,value)			   /*                        */
+static void def_item(name,value)		   /*                        */
   register char * name;				   /*                        */
   register char * value;			   /*                        */
 {						   /*                        */
@@ -233,11 +291,17 @@ void def_item(name,value)			   /*                        */
 
 /*-----------------------------------------------------------------------------
 ** Function:	def_field_type()
-** Purpose:	
+** Purpose:	This function adds a printing representation for a
+**		field name to the used list. The argument is an
+**		equation of the following form
 **		
-**
+**			\textit{type = value}
+**		
+**		\textit{type} is translated to lower case and compared
+**		against the internal representation. \textit{value} is
+**		printed at the appropriate places instead.
 ** Arguments:
-**	s
+**	s	String containing an equation.	
 ** Returns:	nothing
 **___________________________________________________			     */
 void def_field_type(s)				   /*                        */
@@ -250,7 +314,7 @@ void def_field_type(s)				   /*                        */
   while ( is_allowed(*s) ) ++s;			   /*                        */
   c = *s; *s = '\0'; val = new_string(name); *s = c;/*                       */
  						   /*                        */
-  { register char * cp;				   /*                        */
+  { char * cp;				   	   /*                        */
     for ( cp=val; *cp; ++cp ) *cp = ToLower(*cp);  /*                        */
   }						   /*                        */
  						   /*                        */
@@ -288,16 +352,20 @@ static char * get_mapped_or_cased(name,mac,type)   /*                        */
   register Macro mac;				   /*                        */
 { static StringBuffer* sb = (StringBuffer*)NULL;   /*                        */
  						   /*                        */
-  for ( ; mac != MacNull; mac=NextMacro(mac) )	   /*                        */
+  for ( ; mac != MacroNULL; mac=NextMacro(mac) )   /*                        */
   { if ( name == MacroName(mac) )		   /*                        */
-      return MacroVal(mac);			   /*                        */
+      return MacroValue(mac);			   /*                        */
   }						   /*                        */
  						   /*                        */
-  if ( type != SYMBOL_TYPE_LOWER ) 	   	   /*                        */
+  if ( sb == NULL &&				   /*                        */
+       (sb=sbopen()) == NULL )			   /*                        */
+  { OUT_OF_MEMORY("get_item()"); } 		   /*                        */
+  sbrewind(sb);
+  if ( type == SYMBOL_TYPE_LOWER ) 	   	   /*                        */
+  { sbputs(name,sb);
+  }
+  else
   { 						   /*                        */
-    if ( sb == NULL &&				   /*                        */
-	(sb=sbopen()) == NULL )			   /*                        */
-    { OUT_OF_MEMORY("get_item()"); } 		   /*                        */
     						   /*                        */
     switch ( type )			   	   /*                        */
     { case SYMBOL_TYPE_CASED:			   /*                        */
@@ -317,11 +385,9 @@ static char * get_mapped_or_cased(name,mac,type)   /*                        */
 	for ( ; *name; ++name )			   /*                        */
 	{ (void)sbputc(ToUpper(*name),sb); }	   /*                        */
 	break;					   /*                        */
-      }						   /*                        */
-    name = sbflush(sb);				   /*                        */
-    sbrewind(sb);				   /*                        */
+    }						   /*                        */
   }						   /*                        */
-  return name;					   /*                        */
+  return sbflush(sb);				   /*                        */
 }						   /*------------------------*/
 
 
@@ -332,13 +398,17 @@ static char * get_mapped_or_cased(name,mac,type)   /*                        */
 
 /*-----------------------------------------------------------------------------
 ** Function:	get_item()
-** Purpose:	
-**		
-**
+** Purpose:	Return the print representation of a \BibTeX{} string.
+**		The appearance is determined by the |items| mapping.
+**		If no approriate entry is found then |type| is used to
+**		decide whether the item should be returned as
+**		uppercase, lowercase or first upper only.
 ** Arguments:
-**	name
-**	type
-** Returns:	
+**	name	Symbol to get the print representation for.
+**	type	One of the values |SYMBOL_TYPE_UPPER|, |SYMBOL_TYPE_LOWER|, or
+**		|SYMBOL_TYPE_CASED| as they are defined in |type.h|.
+** Returns:	A pointer to a static string. This location  is reused
+**		upon the next invocation of this function.
 **___________________________________________________			     */
 char * get_item(name,type)			   /*                        */
   register char * name;				   /*                        */
@@ -351,7 +421,7 @@ char * get_item(name,type)			   /*                        */
 /***									   ***/
 /*****************************************************************************/
 
- static Macro keys = MacNull;			   /*                        */
+ static Macro keys = MacroNULL;			   /*                        */
 
 /*-----------------------------------------------------------------------------
 ** Function:	save_key()
