@@ -1,32 +1,73 @@
 /******************************************************************************
-** $Id: main.c,v 2.24 1996/02/22 21:43:51 gerd Exp gerd $
+** $Id: main.c,v 2.34 1997/12/06 22:56:00 gerd Exp gerd $
 **=============================================================================
 ** 
 ** This file is part of BibTool.
 ** It is distributed under the GNU General Public License.
 ** See the file COPYING for details.
 ** 
-** (c) 1996 Gerd Neugebauer
+** (c) 1996-1997 Gerd Neugebauer
 ** 
 ** Net: gerd@informatik.uni-koblenz.de
 ** 
+** >>> See the file Doc/c_lib.tex for a printed documentation of the
+** >>> main parts of the C source.
+** 
+**-----------------------------------------------------------------------------
+** Description:
+**	This is the \BibTool{} main module. It contains the |main()|
+**	function which evaluates the command line arguments and
+**	proceeds accordingly. This means that usually resource files
+**	and \BibTeX{} files are read and one or more \BibTeX{} files
+**	are written.
+**
+**	This file makes use of the \BibTool{} C library but is not
+**	part of it. For this purpose it has to provide certain
+**	functions which are expected by the library. These functions
+**	are: 
+**	\begin{quote}
+**	  \texttt{save\_input\_file()}\\
+**	  \texttt{save\_macro\_file()}\\
+**	  \texttt{save\_output\_file()}\\
+**	\end{quote}
+**	The arguments and the expected behaviour of these functions is
+**	described below.
+**
+**	If you are trying to understand the implementation of
+**	\BibTool{} the file |resource.h| plays a central
+**	r\^ole. Consult the description of this file for further
+**	details. 
+**
+**	If you are trying to write your own program to manipulate
+**	\BibTeX{} files then this file can serve as a starting point.
+**	But you should keep in mind that this file has grown over
+**	several years and it contains the full complexity of the
+**	\BibTool{} program logic. Thus you can reduce this file
+**	drastically if you start playing around with the \BibTool{} C
+**	library. 
+**
 ******************************************************************************/
 
-#include "tex_aux.h"
-#include "bibtool.h"
-#include "error.h"
-#include "key.h"
-#include "parse.h"
-#include "pxfile.h"
-#include "print.h"
-#include "rewrite.h"
-#include "rsc.h"
-#include "entry.h"
-#include "macros.h"
-#include "type.h"
-#include "s_parse.h"
-#include "sbuffer.h"
-#include "version.h"
+#include <bibtool/general.h>
+#include <bibtool/init.h>
+#include <bibtool/tex_aux.h>
+#include <bibtool/error.h>
+#include <bibtool/key.h>
+#include <bibtool/pxfile.h>
+#include <bibtool/print.h>
+#include <bibtool/rewrite.h>
+#include <bibtool/rsc.h>
+#include <bibtool/entry.h>
+#include <bibtool/parse.h>
+#include <bibtool/macros.h>
+#include <bibtool/s_parse.h>
+#include <bibtool/symbols.h>
+#include <bibtool/sbuffer.h>
+#include <bibtool/expand.h>
+#ifdef HAVE_LIBKPATHSEA
+#include <kpathsea/debug.h>
+#endif
+#include "config.h"
 
 /*****************************************************************************/
 /* Internal Programs							     */
@@ -38,27 +79,25 @@
 #define _ARG(A) ()
 #endif
  int main _ARG((int argc,char *argv[]));	   /* main.c                 */
- static void keep _ARG((Record rec));		   /* main.c                 */
- static void process_stored_records _ARG((int verbose));/* main.c            */
- static void read_loop _ARG((void (*fct)_ARG((Record))));/* main.c           */
- static void store _ARG((Record rec));		   /* main.c                 */
- static void update_crossref _ARG((Record rec,Record first_rec));/* main.c   */
- void printchar _ARG((char c));			   /* main.c                 */
+ static int dbl_check _ARG((DB db,Record rec));	   /* main.c                 */
+ static int do_keys _ARG((DB db,Record rec));	   /* main.c                 */
+ static int do_no_keys _ARG((DB db,Record rec));   /* main.c                 */
+ static int rec_gt _ARG((Record r1,Record r2));	   /* main.c                 */
+ static int rec_lt _ARG((Record r1,Record r2));	   /* main.c                 */
+ static int update_crossref _ARG((DB db,Record rec));/* main.c               */
+ static void usage _ARG((int fullp));		   /* main.c                 */
  void save_input_file _ARG((char *file));	   /* main.c                 */
- void save_macro_file _ARG((char * file));	   /* main.c                 */
+ void save_macro_file _ARG((char *file));	   /* main.c                 */
  void save_output_file _ARG((char * file));	   /* main.c                 */
- void usage _ARG((int full));			   /* main.c                 */
 
 /*****************************************************************************/
-/* External Programs							     */
+/* External Programs and Variables					     */
 /*===========================================================================*/
 
- extern char * sym_crossref;
-
 /*---------------------------------------------------------------------------*/
 
 #ifndef __STDC__
-#ifndef HAS_getenv
+#ifndef HAVE_GETENV
 
  char * getenv _ARG((char *name));		   /* main.c                 */
 
@@ -66,7 +105,7 @@
 ** Dummy funtion returning NULL to indicate failure.
 **___________________________________________________			     */
 char * getenv(name)				   /*			     */
-  register char *name;				   /*			     */
+  char *name;				   	   /*			     */
 { return (char*)0;				   /*			     */
 }						   /*------------------------*/
 #endif
@@ -110,32 +149,64 @@ char * getenv(name)				   /*			     */
     "\t%c$\t\tSymbol table output (debugging only)\n",
 #endif
     0L,
-    "Gerd Neugebauer $Date: 1996/02/22 21:43:51 $",
+    "Copyright (C) Gerd Neugebauer $Date: 1997/12/06 22:56:00 $",
     "gerd@informatik.uni-koblenz.de"
   };
 
 /*-----------------------------------------------------------------------------
 ** Function:	usage()
-** Purpose:	Print the version number and
-**		a short description of the command line options.
-**
+** Purpose:	Print the version number, a copyright notice, and
+**		a short description of the command line options on the
+**		|stderr| stream.
+**		In addition certain configuration options are
+**		printed. If the argument is |FALSE| then only the
+**		copyright and the version number are displayed.
 ** Arguments:
-**	full	Boolean. If FALSE only the versin is displayed.
+**	fullp	Boolean. If |FALSE| only the version is displayed.
 ** Returns:	nothing
 **___________________________________________________			     */
-void usage(full)				   /*			     */
-  int           full;				   /*                        */
+static void usage(fullp)			   /*			     */
+  int           fullp;				   /*                        */
 { register char **cpp;				   /*			     */
+  static char    *comma = ", ";			   /*                        */
+  char		 *sep   = " ";			   /*                        */
 						   /*			     */
-  ErrPrintF("BibTool Vers. %s (C) 1996 Gerd Neugebauer\n\n",version);/*	     */
+  show_version();				   /*                        */
 						   /*			     */
-  if ( full )					   /*                        */
+  if ( fullp )					   /*                        */
   { for ( cpp=use; *cpp; cpp++ )		   /*			     */
     { ErrPrintF2(*cpp,				   /*			     */
 		 OptionLeadingCharacter,	   /*			     */
 		 OptionLeadingCharacter);	   /*			     */
     }						   /*                        */
   }						   /*                        */
+						   /*                        */
+  ErrPrintF("\nLibrary path: %s",		   /*                        */
+	    (rsc_v_rsc==NULL		   	   /*                        */
+	     ? "none"				   /*                        */
+	     : rsc_v_rsc));	   	   	   /*                        */
+ 						   /*                        */
+  ErrPrint("\nSpecial configuration options:");	   /*                        */
+#define SPECIAL_OPTIONS " none"
+#ifdef HAVE_LIBKPATHSEA
+  ErrPrintF("%skpathsea",sep);			   /*                        */
+  sep = comma;					   /*                        */
+#undef SPECIAL_OPTIONS
+#define SPECIAL_OPTIONS ""
+#endif
+#ifndef REGEX
+  ErrPrintF("%sNO regex",sep);			   /*                        */
+  sep = comma;					   /*                        */
+#undef SPECIAL_OPTIONS
+#define SPECIAL_OPTIONS ""
+#endif
+#ifdef EMTEX_LIKE_PATH
+  ErrPrintF("%semTeX like path",sep);		   /*                        */
+  sep = comma;					   /*                        */
+#undef SPECIAL_OPTIONS
+#define SPECIAL_OPTIONS ""
+#endif
+  ErrPrintF("%s\n",SPECIAL_OPTIONS);		   /*                        */
 }						   /*------------------------*/
 
 
@@ -166,11 +237,14 @@ void usage(full)				   /*			     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	save_input_file()
-** Purpose:	The input file pipe is a dynamic array of strings
+** Purpose:	The input file pipe is a dynamic array of strings.
+**		This fifo stack is used to store the input \BibTeX{}
+**		files to be processed by \BibTool.
+**
 **		This function is called to push an string into the pipe.
 **		If neccesary the array has to be allocated or enlarged.
 **		This is done in larger junks to avoid lots of calls to
-**		realloc().
+**		|realloc()|.
 ** Arguments:
 **	file	File name to save.
 ** Returns:	nothing
@@ -182,7 +256,7 @@ void save_input_file(file)			   /*			     */
   { WARNING("Missing input file name. Flag ignored.");/*		     */
     return;					   /*			     */
   }						   /*			     */
-  if ( *file == '-' && *(file+1) == '\0' )	   /*			     */
+  if ( *file == '-' && file[1] == '\0' )	   /*			     */
   { file = NULL; }				   /*			     */
 						   /*			     */
   if ( InputPipeIsFull )			   /* No space left?	     */
@@ -241,7 +315,7 @@ void save_output_file(file)			   /*			     */
 ** Returns:	nothing
 **___________________________________________________			     */
 void save_macro_file(file)			   /*			     */
-  char * file;					   /*			     */
+  char *file;					   /*			     */
 { if ( macro_file != NULL )			   /*			     */
   { WARNING2("Macro file redefined: ",file); }	   /*			     */
   macro_file = file;				   /*			     */
@@ -256,47 +330,37 @@ void save_macro_file(file)			   /*			     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	main()
-** Purpose:	
-**		Perform Initializations.
-**		Evaluate command line arguments.
-**		Run the main loop.
+** Purpose:	This is the main function which is automatically
+**		called when the program is started. This function
+**		contains the overall program logic. It has to perform
+**		the appropriate initializations, evaluate command line
+**		arguments, and run the main loop.
 ** Arguments:
 **	argc	Number of arguments
 **	argv	Array of arguments
-** Returns:	
+** Returns:	0 upon success. Usually a failure raises an exception
+**		which leads to an |exit()|. Thus this function does
+**		not need to signal an error to the calling environment.
 **___________________________________________________			     */
 int main(argc,argv)				   /*			     */
-  int		argc;				   /* Argument count	     */
-  char		*argv[];			   /* Argument values	     */
-{ register int	i;				   /*			     */
-  register char *ap;				   /*			     */
-  int		need_rsc = TRUE;		   /*			     */
-  void		(*fct)();			   /* Function pointer	     */
-#ifdef EMTEX_LIKE_PATH
-  static StringBuffer *sb_rsc;			   /*                        */
-  static StringBuffer *sb_bibtex;		   /*                        */
+  int	argc;				   	   /* Argument count	     */
+  char	*argv[];			   	   /* Argument values	     */
+{ DB	the_db;				   	   /* The \BibTool{} program */
+						   /* currently handles a    */
+						   /* single database at a   */
+						   /* time.  The local       */
+						   /* variable |the_db|      */
+						   /* contains a reference to*/
+						   /* this database.	     */
+  int	i;				   	   /*			     */
+  char	*ap;				   	   /*			     */
+  FILE	*file;				   	   /*                        */
+  int	need_rsc = TRUE;		   	   /*			     */
+  int	(*fct)();			   	   /* Function pointer	     */
+  int	c_len;					   /*                        */
+  int   *c;					   /*                        */
  						   /*                        */
-  { char *emtexdir = getenv(EMTEXDIR);		   /*                        */
-    if ( emtexdir == NULL ) emtexdir = EMTEXTDIR_DEFAULT;/*                  */
-    sb_rsc = sbopen();				   /*                        */
-    sbputs(".;",sb_rsc);			   /*                        */
-    sbputs(emtexdir,sb_rsc);			   /*                        */
-    sbputs(EMTEXT_RESOURCE,sb_rsc);		   /*                        */
-    rsc_v_rsc = sbflush(sb_rsc);		   /*                        */
-    sb_bibtex = sbopen();			   /*                        */
-    sbputs(".;",sb_bibtex);			   /*                        */
-    sbputs(emtexdir,sb_bibtex);			   /*                        */
-    sbputs(EMTEXT_BIBTEX,sb_bibtex);		   /*                        */
-    rsc_v_bibtex = sbflush(sb_bibtex);		   /*                        */
-  }						   /*                        */
-#endif
-						   /*			     */
-  init_type();					   /*			     */
-  init_symbols();				   /*			     */
-  init_entries();				   /*			     */
-  init_read();					   /*			     */
-  init_key(0);					   /*			     */
-  set_rsc_path(rsc_v_rsc);			   /*                        */
+  init_bibtool(argv[0]);			   /*                        */
 						   /*			     */
   for ( i=1; i<argc; i++ )			   /*			     */
   { if ( *(ap=argv[i]) != OptionLeadingCharacter ) /*			     */
@@ -306,6 +370,9 @@ int main(argc,argv)				   /*			     */
     { switch ( *++ap )				   /*			     */
       { case 'A': set_base(ap+1);	    break; /* disambiguation	     */
 	case 'd': Toggle(rsc_double_check); break; /* double entries	     */
+#ifdef HAVE_LIBKPATHSEA
+	case 'D': kpathsea_debug = atoi(++ap); break;/* kpathsea debugging   */
+#endif
 	case 'f': add_format(argv[++i]);	   /*	!!! no break !!!     */
 	case 'F': rsc_make_key = TRUE;	    break; /* key generation	     */
 	case 'h': usage(TRUE); return 1;	   /* print help	     */
@@ -321,11 +388,12 @@ int main(argc,argv)				   /*			     */
 	case 'q': Toggle(rsc_quiet);	    break; /* quiet		     */
 	case 'r':				   /* resource file	     */
 	  if ( ++i < argc && load_rsc(argv[i]) )   /*			     */
+	  { need_rsc = FALSE; }		   	   /*			     */
+	  else					   /*                        */
 	  {  NoRscError(argv[i]); }		   /*                        */
-	  else need_rsc = FALSE;		   /*			     */
 	  break;				   /*			     */
 	case 'R': need_rsc = search_rsc();  break; /* default resource file  */
-	case 's': Toggle(rsc_sort); break;	   /* sort		     */
+	case 's': Toggle(rsc_sort);	    break; /* sort		     */
 	case 'S':				   /*			     */
 	  Toggle(rsc_sort);			   /*			     */
 	  Toggle(rsc_sort_reverse);		   /*			     */
@@ -335,7 +403,10 @@ int main(argc,argv)				   /*			     */
 	case 'x':				   /* extract		     */
 	  rsc_all_macs = FALSE;			   /*                        */
 	  if ( ++i < argc )			   /*			     */
-	  { read_aux(argv[i],*++ap=='v'); }	   /*			     */
+	  { if ( read_aux(argv[i],save_input_file,*++ap=='v') )/*            */
+	    { ERROR3("aux File ",argv[i]," not found.");/*                   */
+	    }					   /*                        */
+	  }					   /*                        */
 	  else		    { NoSFileWarning; }	   /*			     */
 	  break;				   /*			     */
 	case 'X':				   /* extract pattern	     */
@@ -346,8 +417,8 @@ int main(argc,argv)				   /*			     */
 	case '#': Toggle(rsc_cnt_all);	    break; /* print full statistics  */
 	case '@': Toggle(rsc_cnt_used);	    break; /* print short statistics */
 	case '-':				   /* extended command	     */
-	  if ( *++ap )	      { use_rsc(ap);	  }/*			     */
-	  else if ( ++i<argc ){ use_rsc(argv[i]); }/*			     */
+	  if ( *++ap )	      { (void)use_rsc(ap); }/*			     */
+	  else if ( ++i<argc ){ (void)use_rsc(argv[i]); }/*		     */
 	  else		      { MissingResource;  }/*			     */
 	  break;				   /*			     */
 #ifdef SYMBOL_DUMP
@@ -366,256 +437,176 @@ int main(argc,argv)				   /*			     */
   if ( InputPipeIsEmpty )			   /* If no input file given */
   { save_input_file("-"); }			   /*  then read from stdin  */
 						   /*			     */
-  if ( output_file != NULL )			   /* If output file is given*/
-  { (void)tell(output_file); }			   /*  then use it.	     */
-						   /*			     */
-  if	  (  rsc_sort				   /*			     */
-	  || rsc_sort_reverse ) fct = store;	   /*			     */
-  else if (  rsc_double_check			   /*			     */
-	  || rsc_make_key     ) fct = keep;	   /*			     */
-  else				fct = print_record;/*			     */
-						   /*			     */
-  init_macros();				   /*			     */
-  init_read();					   /*			     */
-  init_key(1);					   /*			     */
+  init_read();					   /* Just in case the path  */
+ 						   /*  has been modified.    */
+  the_db = new_db();				   /*                        */
 						   /*			     */
   for ( i=0; i<input_file_ptr; i++ )		   /* For all input files    */
   {						   /*			     */
-    if ( !see_bib(input_files[i]) )		   /* Try to open bib file   */
+    if ( read_db(the_db,			   /*                        */
+		 input_files[i],		   /*                        */
+		 is_selected,			   /*                        */
+		 rsc_verbose) )	   		   /*                        */
     { NoFileError(input_files[i]); }		   /*			     */
-    else					   /*			     */
-    { if ( rsc_verbose )			   /* If desired print an    */
-      { VerbosePrint2("Reading ",filename); }	   /*	open message.	     */
-						   /*			     */
-      read_loop(fct);				   /* Process the whole file.*/
-						   /*			     */
-      (void)seen();				   /* Close the file.	     */
-						   /*			     */
-      if ( rsc_verbose )			   /* If desired print a     */
-      { VerbosePrint2("Done with ",filename); }	   /*	close message.	     */
-    }						   /*			     */
   }						   /*			     */
+ 						   /*                        */
+  apply_aux(the_db);				   /*                        */
+ 						   /*                        */
+  if ( rsc_cnt_all || rsc_cnt_used )		   /*			     */
+  { int i;
+    int * cnt = db_count(the_db,&c_len);
+
+    if ( (c=(int*)malloc(c_len*sizeof(int))) == NULL )
+    { rsc_cnt_all = rsc_cnt_used = 0; }
+    else
+    { for (i=0;i<c_len;i++) c[i] = cnt[i];
+    }
+  }
 						   /*			     */
-  process_stored_records(rsc_verbose);		   /* Maybe something's left.*/
+  if ( rsc_make_key )				   /*                        */
+  { if ( rsc_key_preserve )	   		   /*                        */
+    { db_forall(the_db,mark_key); }		   /*                        */
+    db_rewind(the_db);				   /*                        */
+    db_forall(the_db,do_keys);		   	   /*                        */
+    db_forall(the_db,update_crossref);	   	   /*                        */
+  }						   /*                        */
+  else						   /*                        */
+  {						   /*                        */
+    db_forall(the_db,do_no_keys);		   /*                        */
+  }						   /*                        */
+ 						   /*                        */
+  if ( rsc_sort ) {				   /*                        */
+    if ( rsc_sort_reverse ) fct = rec_lt;	   /*                        */
+    else		    fct = rec_gt;	   /*                        */
+    db_sort(the_db,fct);			   /*                        */
+  }						   /*                        */
+ 						   /*                        */
+  if ( rsc_srt_macs )		   	   	   /* Maybe sort macros      */
+  { db_mac_sort(the_db); }			   /*                        */
+ 						   /*                        */
+  if ( rsc_double_check	)		   	   /* Maybe look for doubles */
+  { db_forall(the_db,dbl_check); }		   /*                        */
+ 						   /*                        */
+  if ( output_file == NULL ||			   /*                        */
+       (file=fopen(output_file,"w")) == NULL )	   /*                        */
+  { file = stdout; }				   /*                        */
+  print_db(file,the_db,rsc_print_et);		   /*                        */
+  if ( file != stdout ) { fclose(file); }	   /*                        */
 						   /*			     */
-  (void)told();					   /* Close output stream.   */
-						   /*			     */
-  if ( macro_file != NULL )			   /*			     */
+  if ( macro_file != NULL  && *macro_file )	   /*			     */
   { if ( rsc_verbose )				   /*			     */
     { VerbosePrint1("Writing macros"); }	   /*			     */
-    dump_mac(macro_file,rsc_all_macs);		   /* Write the strings.     */
+ 						   /*                        */
+    if ( strcmp(macro_file,"-") == 0 ||		   /*                        */
+	 (file=fopen(macro_file,"w")) == NULL )    /*                        */
+    { file = stdout; }				   /*                        */
+    print_db(file,the_db,rsc_all_macs?"s":"S");	   /*                        */
+    if ( file != stdout ) { fclose(file); }	   /*                        */
   }						   /*			     */
 						   /*			     */
   if ( rsc_cnt_all || rsc_cnt_used )		   /*			     */
-  { entry_statistics(rsc_cnt_all); }		   /* Write statistics.	     */
+  { int i;					   /*                        */
+    int *cnt = db_count(the_db,NULL);		   /*                        */
+						   /*                        */
+    (void)fputc('\n',err_file);		   	   /*			     */
+    for (i=0; i<c_len; ++i)			   /*			     */
+    { if ( rsc_cnt_all || c[i] > 0 )	   	   /*			     */
+      { (void)fprintf(err_file,			   /*			     */
+		      "---  %-15s %5d read  %5d written\n",/*		     */
+		      get_entry_type(i),	   /*			     */
+		      c[i],		   	   /*			     */
+		      cnt[i]);		   	   /*			     */
+      }						   /*			     */
+    }						   /*			     */
+  }						   /*                        */
 						   /*			     */
 #ifdef SYMBOL_DUMP
   if ( rsc_dump_symbols ) sym_dump();		   /* Write symbols.	     */
+#endif
+ 						   /*                        */
+#ifdef FREE_MEMORY
+  free_db(the_db);				   /*                        */
+  the_db = NULL;				   /*                        */
+  sym_gc();					   /*                        */
+#ifdef SYMBOL_DUMP
+  sym_dump();					   /*                        */
+#endif
 #endif
   return 0;					   /*			     */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:	read_loop()
-** Purpose:	Read one file until EOF and do what's to be done.
-** Arguments:
-**	fct
-** Returns:	nothing
-**___________________________________________________			     */
-static void read_loop(fct)			   /*			     */
-  void		  (*fct)_ARG((Record));		   /* Function pointer	     */
-{ register int	  type;				   /*			     */
-  Record rec;				   	   /*			     */
-						   /*			     */
-  while ( (type=parse()) != EndOfFile )		   /*			     */
-  { if ( type < 0 )				   /* Errors give rise to    */
-    { SkipWarning; }				   /*  a warning.	     */
-    else if ( IsSpecialRecord(type) )		   /* STRING/PREAMBLE/COMMENT*/
-    { print_record(get_master_record());	   /* are written at once    */
-      ++EntryUsed(type);			   /*			     */
-    }						   /*			     */
-    else					   /*			     */
-    {						   /*			     */
-      rec = get_master_record();		   /*			     */
-						   /*			     */
-      if ( rsc_make_key ) make_key(rec);	   /* Maybe make a new key   */
-      else		  mark_key(rec);	   /*  or remember the old   */
-						   /*			     */
-      if (   rsc_select	!= TRUE			   /* selection enabled	and  */
-	  || lookup_item(rec) )	   		   /* record selected	     */
-      {						   /*  then		     */
-	++EntryUsed(type);			   /*			     */
-	rewrite_record(rec);			   /*			     */
-	(*fct)(rec);				   /* Do what's needed to be */
-						   /*  done.		     */
-	if ( rsc_verbose ) { ErrC('+'); FlushErr; }/*			     */
-      }						   /*			     */
-      else if ( rsc_verbose )			   /*			     */
-      { ErrC('.'); FlushErr; }			   /* Just to see something. */
-    }						   /*			     */
-  }						   /*			     */
-}						   /*------------------------*/
-
-
-/*****************************************************************************/
-/***			      Sorting Section				   ***/
-/*****************************************************************************/
-
- static Record sort_rec = RecNULL;
-
-#define RecLess(A,B)							\
-	((rsc_sort_reverse? -strcmp(RecordKey(A),RecordKey(B))		\
-			  :  strcmp(RecordKey(A),RecordKey(B))) < 0)
-
-/*-----------------------------------------------------------------------------
-** The sorting uses a variant of insertion sort.
-** Thus the major task is performed in this function, which is called to store
-** a record.
-** The Record structure contains two pointers to the same type. With those
-** pointers a double linked list is established. sort_rec is either NULL ---
-** if no record is stored --- or it is a pointer to some element in the
-** ORDERED list:
-** 
-**    +------+	 +------+   +------+   +------+	  +------+   +------+	
-** -->|Record|-->|Record|-->|Record|-->|Record|-->|Record|-->|Record|-->
-** <--|	     |<--|	|<--|	   |<--|      |<--|	 |<--|	    |<--
-**    +------+	 +------+   +------+   +------+	  +------+   +------+	
-**			       ^
-**		sort_rec_______|
-** 
-** With this picture in mind it's obvious what's to be done.
-** There are two major cases. To find the right position to insert the new
-** record
-** - either the pointer has to be moved leftward
-** - or	    the pointer has to be moved rightward
-** 
-** Special cases have to be considered if the new record is to be inserted
-** before the first or after the last element in the list.
-** 
-** The rest is simply legwork.
+** Function:	rec_gt()
+** Purpose:	
+**		
 **
-** Well, a word on the complexity.
-** If the input is almost sorted (correct or in reverse order) then the 
-** algorithm is linear.
-** In the worst case the algorithm is quadratic. This worst case looks as
-** follows:
-** A record has to be inserted which is smaller than the least element in the
-** list. Afterwards one which is greater than the largest.
-** 
-** NOTE: CURRENTLY THE ALGORITHM IS NOT STABLE:
-**	 Records with the same key may get mixed in the sorted list.
+** Arguments:
+**	r1
+**	r2
+** Returns:	
 **___________________________________________________			     */
-static void store(rec)				   /*			     */
-  register Record rec;				   /*			     */
-{ register Record rp;				   /*			     */
-						   /*			     */
-  rec = copy_record(rec);			   /* Make a private copy.   */
-						   /*			     */
-  make_sort_key(rec);				   /*			     */
-						   /*			     */
-  if ( sort_rec == RecNULL )			   /* List is empty	     */
-  { sort_rec = rec; return; }			   /* Just remember the rec. */
-						   /*			     */
-  rp	   = sort_rec;				   /*			     */
-  sort_rec = rec;				   /*			     */
-  if ( RecLess(rec,rp) )			   /*			     */
-  { while ( PrevRecord(rp) != RecNULL )		   /* Move leftward.	     */
-    { rp = PrevRecord(rp);			   /*			     */
-      if ( !RecLess(rec,rp) )			   /*			     */
-      { PrevRecord(rec)		   = rp;	   /* Insert		     */
-	NextRecord(rec)		   = NextRecord(rp);/*			     */
-	PrevRecord(NextRecord(rec))= rec;	   /*			     */
-	NextRecord(rp)		   = rec;	   /*			     */
-	return;					   /*			     */
-      }						   /*			     */
-    }						   /*			     */
-    PrevRecord(rp)  = rec;			   /* Insert before the first*/
-    NextRecord(rec) = rp;			   /*			     */
-  }						   /*			     */
-  else						   /*			     */
-  { while ( NextRecord(rp) != RecNULL )		   /* Move rightward.	     */
-    { rp = NextRecord(rp);			   /*			     */
-      if ( RecLess(rec,rp) )			   /*			     */
-      { NextRecord(rec)		   = rp;	   /* Insert		     */
-	PrevRecord(rec)		   = PrevRecord(rp);/*			     */
-	NextRecord(PrevRecord(rec))= rec;	   /*			     */
-	PrevRecord(rp)		   = rec;	   /*			     */
-	return;					   /*			     */
-      }						   /*			     */
-    }						   /*			     */
-    NextRecord(rp)  = rec;			   /* Insert after the last. */
-    PrevRecord(rec) = rp;			   /*			     */
-  }						   /*			     */
+static int rec_gt(r1,r2)			   /*                        */
+  Record r1;					   /*                        */
+  Record r2;					   /*                        */
+{ return (strcmp(RecordSortkey(r1),RecordSortkey(r2)) < 0);/*                */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:	keep()
-** Purpose:	Just push a copy of the given record to the list.
+** Function:	rec_lt()
+** Purpose:	
+**		
+**
 ** Arguments:
-**	rec	Record to push
-** Returns:	nothing
+**	r1
+**	r2
+** Returns:	
 **___________________________________________________			     */
-static void keep(rec)				   /*			     */
-  register Record rec;				   /*			     */
-{						   /*			     */
-  rec = copy_record(rec);			   /* Make a private copy.   */
-						   /*			     */
-  make_sort_key(rec);				   /*			     */
-						   /*			     */
-  if ( sort_rec == RecNULL )			   /* List is empty	     */
-  { sort_rec = rec; }				   /* Just remember the rec. */
-  else						   /*			     */
-  { NextRecord(sort_rec) = rec;			   /*			     */
-    PrevRecord(rec)	 = sort_rec;		   /*			     */
-    sort_rec		 = rec;			   /*			     */
-  }						   /*			     */
+static int rec_lt(r1,r2)			   /*                        */
+  Record r1;					   /*                        */
+  Record r2;					   /*                        */
+{ return (strcmp(RecordSortkey(r1),RecordSortkey(r2)) > 0);/*                */
 }						   /*------------------------*/
 
-
-#define equal_records(R1,R2) RecordKey(R1) == RecordKey(R2)
+/*-----------------------------------------------------------------------------
+** Function:	do_keys()
+** Purpose:	
+**		
+**
+** Arguments:
+**	rec
+** Returns:	
+**___________________________________________________			     */
+static int do_keys(db,rec)			   /*                        */
+  DB	 db;					   /*                        */
+  Record rec;					   /*                        */
+{						   /*                        */
+  rewrite_record(db,rec);			   /*			     */
+  sort_record(rec);				   /*                        */
+  make_key(db,rec);				   /*                        */
+  if ( rsc_sort || rsc_double_check )		   /*                        */
+  { make_sort_key(db,rec); }		   	   /*                        */
+  return 0;					   /*                        */
+}						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:	process_stored_records()
-** Purpose:	This function performs actions for sorting after all records
-**		have been read.
-**		It simply goes to the beginning and prints them all.
+** Function:	do_no_keys()
+** Purpose:	
+**		
+**
 ** Arguments:
-**	verbose	boolean
-** Returns:	nothing
+**	rec
+** Returns:	
 **___________________________________________________			     */
-static void process_stored_records(verbose)	   /*			     */
-  int		  verbose;			   /*			     */
-{ register Record rec;				   /*			     */
-  Record	  first_rec;			   /*			     */
-						   /*			     */
-  if ( (rec=sort_rec) == RecNULL ) return;	   /* No records to process. */
-						   /*			     */
-  if ( verbose ) { VerbosePrint1("Writing records"); }/*		     */
-						   /*			     */
-  while( PrevRecord(rec) != RecNULL )		   /*			     */
-    rec = PrevRecord(rec);			   /* rewind		     */
-						   /*			     */
-  first_rec = rec;				   /*			     */
-						   /*			     */
-  print_record(rec);				   /* print the first	     */
-  rec = NextRecord(rec);			   /*			     */
-						   /*			     */
-  while ( rec != RecNULL )			   /* loop towards end	     */
-  {						   /*  and print it.	     */
-    if (   IsRecordXREF(rec)			   /*			     */
-	&& rsc_make_key	      )			   /*			     */
-    { update_crossref(rec,first_rec); }		   /*			     */
-						   /*			     */
-    if (   rsc_double_check			   /*			     */
-	&& equal_records(PrevRecord(rec),rec) )	   /*			     */
-    { print_commented(rec);			   /*			     */
-      WARNING3("Possible double entries discovered: \n***\t\t",/*	     */
-	       RecordKey(PrevRecord(rec)),	   /*			     */
-	       RecordKey(rec));			   /*			     */
-    }						   /*			     */
-    else { print_record(rec); }			   /*			     */
-    rec = NextRecord(rec);			   /*			     */
-  }						   /*			     */
+static int do_no_keys(db,rec)			   /*                        */
+  DB	 db;					   /*                        */
+  Record rec;					   /*                        */
+{						   /*                        */
+  rewrite_record(db,rec);			   /*			     */
+  sort_record(rec);				   /*                        */
+  mark_key(db,rec);				   /*                        */
+  if ( rsc_sort || rsc_double_check )		   /*                        */
+  { make_sort_key(db,rec); }		   	   /*                        */
+  return 0;					   /*                        */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
@@ -628,56 +619,78 @@ static void process_stored_records(verbose)	   /*			     */
 **	first_rec
 ** Returns:	nothing
 **___________________________________________________			     */
-static void update_crossref(rec,first_rec)	   /*			     */
-  Record	  rec;				   /*			     */
-  register Record first_rec;			   /*			     */
-{ register char	  **hp;				   /*			     */
-  int		  i;				   /*			     */
-  char		  *ref, *s;			   /*			     */
+static int update_crossref(db,rec)		   /*			     */
+  DB		db;				   /*                        */
+  Record	rec;				   /*			     */
+{ register char	**hp;				   /*			     */
+  register int  i;				   /*                        */
+  char		*t,*s;			   	   /*			     */
 						   /*			     */
+  if ( !RecordIsXREF(rec) ) return 0;		   /*			     */
+						   /*                        */
   for ( i=RecordFree(rec), hp = RecordHeap(rec);   /* search crossref field  */
 	i>0 && *hp != sym_crossref;		   /*			     */
-	i-=2, hp += 2	 )			   /*			     */
+	i-=2, hp += 2 )			   	   /*			     */
   { }						   /*			     */
-						   /*			     */
   if ( i <= 0 )					   /*			     */
   { DebugPrint1("*** No crossref found.");	   /*			     */
-    return;					   /*			     */
+    return 0;					   /*			     */
   }						   /*			     */
 						   /*			     */
-  s = *++hp; s++;				   /*			     */
-  (void)sp_open(s);				   /* Try to extract	     */
-  if ( (ref = SParseSymbol(&s)) == NULL )	   /*  the crossref as symbol*/
-  { return; }					   /*			     */
+  t = *++hp; t++;				   /*			     */
+  (void)sp_open(t);				   /* Try to extract	     */
+  if ( (s = SParseSymbol(&t)) == NULL )	   	   /*  the crossref as symbol*/
+  { return 0; }					   /*			     */
 						   /*			     */
-  while (   first_rec != RecNULL		   /* Search the referenced  */
-	 && RecordOldKey(first_rec) != ref )	   /*  record.		     */
-  { first_rec = NextRecord(first_rec); }	   /*			     */
-						   /*			     */
-  if ( first_rec == RecNULL )			   /*			     */
-  { ERROR2("Crossref not found:",ref);		   /*			     */
-    return;					   /*			     */
+  if ( (s = db_new_key(db,s)) == NULL )	   	   /*			     */
+  { ERROR2("Crossref not found: ",s);		   /*			     */
+    return 0;					   /*			     */
   }						   /*			     */
-  ref = *RecordHeap(first_rec);			   /* Get the new key.	     */
-						   /*			     */
-  if ( (s = malloc(strlen(ref)+3)) == NULL )	   /* allocate temp memory   */
+  if ( (t = malloc(strlen(s)+3)) == NULL )	   /* allocate temp memory   */
   { OUT_OF_MEMORY("update_crossref()"); }	   /*		             */
 						   /*			     */
-  (void)sprintf(s,				   /* construct new crossref */
-		(**hp=='"'?"\"%s\"":"{%s}"),	   /*			     */
-		ref);				   /*			     */
-						   /*			     */
-  *hp = symbol(s);				   /* store new crossref     */
-  free(s);					   /* free temp memory	     */
+  (void)sprintf(t,(**hp=='"'?"\"%s\"":"{%s}"),s);  /* construct new crossref */
+  *hp = symbol(t);				   /* store new crossref     */
+  free(t);					   /* free temp memory	     */
+  return 0;					   /*                        */
+}						   /*------------------------*/
+
+
+#define equal_records(R1,R2) RecordSortkey(R1) == RecordSortkey(R2)
+
+/*-----------------------------------------------------------------------------
+** Function:	dbl_check()
+** Purpose:	
+**		
+**
+** Arguments:
+**	rec
+** Returns:	
+**___________________________________________________			     */
+static int dbl_check(db,rec)			   /*                        */
+  DB	 db;					   /*                        */
+  Record rec;					   /*                        */
+{						   /*                        */
+  if ( PrevRecord(rec) != RecordNULL		   /*                        */
+       && equal_records(PrevRecord(rec),rec) )	   /*			     */
+  { WARNING2("Possible double entries discovered: \n***\t\t",/*	             */
+	     RecordSortkey(rec));		   /*			     */
+    if (rsc_del_dbl)				   /*                        */
+    { delete_record(db,rec); }		   	   /*                        */
+    else 					   /*                        */
+    { SetRecordDELETED(rec); }			   /*                        */
+  }						   /*			     */
+ 						   /*                        */
+  return 0;					   /*                        */
 }						   /*------------------------*/
 
 
 #ifdef DEBUG
 /*-----------------------------------------------------------------------------
-** Function:	printchar()
+** Function*:	printchar()
 ** Purpose:	For debugging of regex.c
 ** Arguments:
-**	c
+**	c	Character to print.
 ** Returns:	nothing
 **___________________________________________________			     */
 void printchar(c)				   /*                        */
