@@ -1,26 +1,24 @@
 /******************************************************************************
-** $Id: names.c,v 0.00 1995/01/ 1 19:47:34 gerd Exp $
+** $Id: names.c,v 1.6 1997/12/03 11:50:30 gerd Exp $
 *******************************************************************************
 ** 
 ** This file is part of BibTool.
 ** It is distributed under the GNU General Public License.
 ** See the file COPYING for details.
 ** 
-** (c) 1996 Gerd Neugebauer
+** (c) 1996-1997 Gerd Neugebauer
 ** 
 ** Net: gerd@informatik.uni-koblenz.de
 ** 
 ******************************************************************************/
 
-#include <stdio.h>
-#include "bibtool.h"
-#include "names.h"
-#include "error.h"
-#include "symbols.h"
-#include "stack.h"
-#include "sbuffer.h"
-#include "s_parse.h"
-#include "type.h"
+#include <bibtool/general.h>
+#include <bibtool/names.h>
+#include <bibtool/error.h>
+#include <bibtool/symbols.h>
+#include <bibtool/stack.h>
+#include <bibtool/sbuffer.h>
+#include <bibtool/s_parse.h>
 
 /*****************************************************************************/
 /* Internal Programs                                                         */
@@ -40,8 +38,10 @@
  static void free_name_node _ARG((NameNode node)); /* names.c                */
  static void initial _ARG((char *s,char *trans,int len,StringBuffer *sb));/* names.c*/
  static void pp_one_name _ARG((StringBuffer *sb,char **w,NameNode format,char *trans,int len,char *comma,int commas));/* names.c*/
+#ifdef BIBTEX_SYNTAX
  static void set_type _ARG((char **sp,char **midp));/* names.c               */
  void set_name_format _ARG((NameNode *nodep,char *s));/* names.c             */
+#endif
 
 /*****************************************************************************/
 /* External Programs                                                         */
@@ -54,14 +54,15 @@
 ** Function:	new_name_node()
 ** Purpose:	Allocate a name node structure and fill it with the values 
 **		given as arguments.
-**
+**		If no more memory is available then en error is raised and
+**		the program is terminated.
 ** Arguments:
 **	type	
 **	strip	
 **	pre	
 **	mid	
 **	post	
-** Returns:	
+** Returns:	The new instance of a node.
 **___________________________________________________			     */
 static NameNode new_name_node(type,strip,pre,mid,post)/*                     */
   int      type;				   /*                        */
@@ -84,11 +85,11 @@ static NameNode new_name_node(type,strip,pre,mid,post)/*                     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	free_name_node()
-** Purpose:	
-**		
-**
+** Purpose:	Release the memory allocated for a node together with all
+**		nodes reachable via a |NextName| chain.
+**		It is not save to access a freed node.
 ** Arguments:
-**	node
+**	node	Node to free.
 ** Returns:	nothing
 **___________________________________________________			     */
 static void free_name_node(node)		   /*                        */
@@ -102,6 +103,45 @@ static void free_name_node(node)		   /*                        */
   }						   /*                        */
 }						   /*------------------------*/
 
+#ifdef DEBUG
+/*-----------------------------------------------------------------------------
+** Function:	dump_name_node()
+** Purpose:	
+**		
+**		
+** Arguments:
+**	n	
+** Returns:	nothing
+**___________________________________________________			     */
+static void dump_name_node(n)			   /*                        */
+  NameNode n;					   /*                        */
+{						   /*                        */
+  fputc('%',err_file);				   /*                        */
+  switch ( NameType(n) & NameTranslationMask )	   /*                        */
+  { case NameLower:    fputc('-',err_file); break; /*                        */
+    case NameUpper:    fputc('+',err_file); break; /*                        */
+    case NameExternal: fputc('*',err_file); break; /*                        */
+  }						   /*                        */
+  if ( NameStrip(n) >= 0 )			   /*                        */
+  { fprintf(err_file,"%d",NameStrip(n)); }	   /*                        */
+  switch ( NameType(n) & NameTypeMask)		   /*                        */
+  { case NameFirst:  fputc('f',err_file); break;   /*                        */
+    case NameLast:   fputc('l',err_file); break;   /*                        */
+    case NameVon:    fputc('v',err_file); break;   /*                        */
+    case NameJr:     fputc('j',err_file); break;   /*                        */
+    case NameString: fputc('s',err_file); break;   /*                        */
+    default:	     fputc('?',err_file); break;   /*                        */
+  }						   /*                        */
+  fprintf(err_file,				   /*                        */
+	  "[%s][%s][%s]\n",			   /*                        */
+	  (NamePre(n)?NamePre(n):""),		   /*                        */
+	  (NameMid(n)?NameMid(n):""),		   /*                        */
+	  (NamePost(n)?NamePost(n):""));	   /*                        */
+}						   /*------------------------*/
+#endif
+
+
+
 #ifdef BIBTEX_SYNTAX
 /*-----------------------------------------------------------------------------
 ** Function:	set_type()
@@ -206,19 +246,10 @@ void set_name_format(nodep,s)			   /*                        */
       *(s-1) = '\0';				   /*                        */
       post = symbol(post);			   /*                        */
       *(s-1) = c;				   /*                        */
-#ifdef DEBUG
-      switch ( type )
-      { case NameFirst:	fprintf(err_file,"f "); break;
-        case NameLast:	fprintf(err_file,"l "); break;
-        case NameVon:	fprintf(err_file,"v "); break;
-        case NameJr: 	fprintf(err_file,"j "); break;
-        default:	fprintf(err_file,"??"); break;
-      }
-      fprintf(err_file,				   /*                        */
-	      "(%d) pre = \"%s\"  mid = \"%s\"  post = \"%s\"\n",/*          */
-	     strip,pre,mid,post);		   /*                        */
-#endif
       *nodep = new_name_node(type,strip,pre,mid,post);/*                     */
+#ifdef DEBUG
+      dump_name_node(*nodep);			   /*                        */
+#endif
       nodep  = &NextName(*nodep);		   /*                        */
     }						   /*                        */
   }						   /*                        */
@@ -246,37 +277,44 @@ NameNode name_format(s)				   /*                        */
   char *cp;					   /*                        */
   NameNode node = NameNULL;			   /*                        */
   NameNode *nnp;				   /*                        */
-#define OptionalArg(S)					\
-  if (*cp=='[')						\
-  { for ( S=++cp; *cp && *cp!=']'; cp++) ;		\
-    c = *cp;						\
-    *cp = '\0';						\
-    if ( *cp ) { S = symbol(S); } else { S = empty; }	\
-    *cp = c;						\
-    if ( *cp ) cp++;					\
-  } else { S = empty; }
+  static char *msg = "Missing ']' in format string.";
+#define OptionalArg(S)						\
+  if ( *cp == '[' )						\
+  { for ( S=++cp; *cp && *cp != ']'; cp++) ;			\
+    if ( *cp )							\
+    { c = *cp; *cp = '\0'; S = symbol(S); *(cp++) = c; }	\
+    else { S = sym_empty; WARNING(msg); }			\
+  } else { S = sym_empty; }
  						   /*                        */
   nnp = &node;					   /*                        */
-  cp = s;					   /*                        */
+  cp  = s;					   /*                        */
   while (*cp)					   /*                        */
   {						   /*                        */
     if ( *cp == '%' )				   /*                        */
-    { strip = 0;				   /*                        */
-      cp++;					   /*                        */
+    {
+      switch ( *++cp )
+      { case '+': type = NameUpper;    cp++; break;
+        case '-': type = NameLower;    cp++; break;
+        case '*': type = NameExternal; cp++; break;
+	default:  type = NameNone;
+      }
       if ( is_digit(*cp) )			   /*                        */
       { strip = 0;				   /*                        */
         while(is_digit(*cp)) { strip = strip*10 + (*cp++)-'0';}/*            */
       } else { strip = -1; }			   /*                        */
  						   /*                        */
       switch ( *cp )				   /*                        */
-      { case 'f': type = NameFirst; break;	   /*                        */
-        case 'l': type = NameLast; break;	   /*                        */
-        case 'v': type = NameVon; break;	   /*                        */
-        case 'j': type = NameJr; break;		   /*                        */
-        default:  type = NoName;		   /*                        */
+      { case 'f': type |= NameFirst; break;	   /*                        */
+        case 'l': type |= NameLast;  break;	   /*                        */
+        case 'v': type |= NameVon;   break;	   /*                        */
+        case 'j': type |= NameJr;    break;	   /*                        */
+        default:  type |= NoName;		   /*                        */
 	  error(ERR_ERROR|ERR_POINT,		   /*                        */
 		"Wrong type of format.",	   /*                        */
-		(char*)0,(char*)0,s,cp,0,(char*)0);/*                        */
+		(Uchar*)0,			   /*                        */
+		(Uchar*)0,			   /*                        */
+		s,				   /*                        */
+		cp,0,(char*)0);			   /*                        */
       }						   /*                        */
       cp++;					   /*                        */
       OptionalArg(pre);				   /*                        */
@@ -285,8 +323,8 @@ NameNode name_format(s)				   /*                        */
     }						   /*                        */
     else 					   /*                        */
     { mid = cp;					   /*                        */
-      while(*cp && *cp!='%') {cp++;}		   /*                        */
-      c = *cp;					   /*                        */
+      while ( *cp && *cp != '%' ) { cp++; }	   /*                        */
+      c   = *cp;				   /*                        */
       *cp = '\0';				   /*                        */
       mid = symbol(mid);			   /*                        */
       *cp = c;					   /*                        */
@@ -299,21 +337,6 @@ NameNode name_format(s)				   /*                        */
     { *nnp = new_name_node(type,strip,pre,mid,post);/*                       */
       nnp = &NextName(*nnp);			   /*                        */
     }						   /*                        */
-#ifdef DEBUG
-    if ( strip > 0 )				   /*                        */
-    { fprintf(err_file,"%%%d",strip);		   /*                        */
-    } else { fprintf(err_file,"%%");		   /*                        */
-    }						   /*                        */
-    switch ( type )				   /*                        */
-    { case NameFirst:  c = 'f'; break;		   /*                        */
-      case NameLast:   c = 'l'; break;		   /*                        */
-      case NameVon:    c = 'v'; break;		   /*                        */
-      case NameJr:     c = 'j'; break;		   /*                        */
-      case NameString: c = 's'; break;		   /*                        */
-      default:	       c = '?'; break;		   /*                        */
-    }						   /*                        */
-    fprintf(err_file,"%c[%s][%s][%s]\n",c,pre,mid,post);/*                   */
-#endif
   }						   /*                        */
   return node;					   /*                        */
 }						   /*------------------------*/
@@ -330,7 +353,8 @@ NameNode name_format(s)				   /*                        */
 **	and	name separator
 **	namesep
 **	etal
-** Returns:	
+** Returns:	Pointer to static string which is reused upon the next
+**		invocation of this function.
 **___________________________________________________			     */
 char * pp_list_of_names(wa,format,trans,max,comma,and,namesep,etal)/*        */
   char     **wa;				   /*                        */
@@ -406,6 +430,7 @@ static void pp_one_name(sb,w,format,trans,len,comma,commas)/*                */
 { NameNode     nn;				   /*                        */
   char         t;				   /*                        */
   char         *type;				   /*                        */
+  char         *tr;				   /*                        */
   int          i,j,again;			   /*                        */
   int	       first,last,von,jr;		   /*                        */
  						   /*                        */
@@ -471,24 +496,33 @@ static void pp_one_name(sb,w,format,trans,len,comma,commas)/*                */
     { type[i++] = 'v'; von++; }			   /*                        */
   }						   /*                        */
  						   /*                        */
-#ifdef DEBUG
-  for (i=0;i<len;i++)
-  { printf("%c ",type[i]);
-  }
+#ifdef NAMES_DEBUG
+  for (i=0;i<len;i++) { printf("%c ",type[i]); }
   puts("");
 #endif
  						   /*                        */
-  for ( nn=format;				   /*                        */
+  for ( nn = format;				   /*                        */
         nn != NameNULL;				   /*                        */
-        nn=NextName(nn) )			   /*                        */
-  { switch(NameType(nn))	   		   /*                        */
+        nn = NextName(nn) )			   /*                        */
+  { int strip = NameStrip(nn);
+#ifdef DEBUG
+    dump_name_node(nn);
+#endif
+    switch( NameType(nn) & NameTypeMask )	   /*                        */
     { case NameFirst:	t = 'f'; j = first; break; /*                        */
       case NameLast:	t = 'l'; j = last;  break; /*                        */
       case NameVon:	t = 'v'; j = von;   break; /*                        */
       case NameJr:	t = 'j'; j = jr;    break; /*                        */
-      case NameString:	j = 0; sbputs(NameMid(nn),sb); break;/*              */
-      default:		t = ' '; j = 0;     break; /*                        */
+      case NameString:	sbputs(NameMid(nn),sb);	   /*                        */
+      default:		t = ' '; j = 0; 	   /*                        */
     }						   /*                        */
+    switch( NameType(nn) & NameTypeMask )	   /*                        */
+    { case NameLower:   tr = trans_lower;   break; /*                        */
+      case NameUpper:   tr = trans_upper;   break; /*                        */
+      case NameNone:    tr = trans_id;      break; /*                        */
+      default:	        tr = trans; 	   	   /*                        */
+    }						   /*                        */
+ 						   /*                        */
     if ( j > 0 )				   /*                        */
     { sbputs(NamePre(nn),sb);			   /*                        */
       again = FALSE;				   /*                        */
@@ -497,7 +531,7 @@ static void pp_one_name(sb,w,format,trans,len,comma,commas)/*                */
 	{ if ( again ) sbputs(NameMid(nn),sb);	   /*                        */
 	  else again = TRUE;			   /*                        */
  						   /*                        */
-	  initial(w[i],trans,NameStrip(nn),sb);	   /*                        */
+	  initial(w[i],tr,strip,sb);	   	   /*                        */
 	}					   /*                        */
       }						   /*                        */
       sbputs(NamePost(nn),sb);			   /*                        */
@@ -558,6 +592,9 @@ static int is_jr(s)				   /*                        */
       return ( case_cmp(s,"en")   ||		   /*                        */
 	       case_cmp(s,"en.")  ||	   	   /*                        */
 	       case_cmp(s,"enior") );	   	   /*                        */
+    case 'm':					   /*                        */
+      s++;					   /*                        */
+      return ( case_cmp(s,"d") );	   	   /*                        */
     case 'p':					   /*                        */
       s++;					   /*                        */
       return ( case_cmp(s,"hD")   ||		   /*                        */
@@ -633,11 +670,11 @@ static int is_lower_word(s)			   /*                        */
 #ifdef STANDALONE
 
 /*-----------------------------------------------------------------------------
-** Function:	pp_names()
+** Function*:	pp_names()
 ** Purpose:	Take a string consisting of a list of names, parse it and
 **		reformat it according to a format specification.
 **		This function is modelled according to the specification
-**		for format.name$ in BibTeX.
+**		for |format.name$| in \BibTeX.
 ** Arguments:
 **	s	string to be parsed and reformatted.
 **	format	
@@ -721,7 +758,7 @@ char * pp_names(s,format,trans,max,namesep,etal)   /*                        */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:    main()
+** Function*:   main()
 ** Purpose:     Main rountine
 ** Arguments:
 **	argc	Number of arguments (+1)
