@@ -1,27 +1,30 @@
 /******************************************************************************
-** $Id: parse.c,v 2.26 1996/02/22 21:43:51 gerd Exp gerd $
+** $Id: parse.c,v 2.34 1997/12/06 22:56:00 gerd Exp gerd $
 **=============================================================================
 ** 
 ** This file is part of BibTool.
 ** It is distributed under the GNU General Public License.
 ** See the file COPYING for details.
 ** 
-** (c) 1996 Gerd Neugebauer
+** (c) 1996-1997 Gerd Neugebauer
 ** 
 ** Net: gerd@informatik.uni-koblenz.de
 ** 
 ******************************************************************************/
 
-#include "bibtool.h"
-#include "type.h"
-#include "entry.h"
-#include "error.h"
-#include "pxfile.h"
-#include "rsc.h"
-#include "stack.h"
-#include "sbuffer.h"
-#include "macros.h"
-#include "print.h"
+#include <bibtool/general.h>
+#include <bibtool/symbols.h>
+#include <bibtool/entry.h>
+#include <bibtool/error.h>
+#include <bibtool/pxfile.h>
+#include <bibtool/rsc.h>
+#include <bibtool/stack.h>
+#include <bibtool/sbuffer.h>
+#include <bibtool/macros.h>
+#include <bibtool/print.h>
+#ifdef HAVE_LIBKPATHSEA
+#include <kpathsea/tex-file.h>
+#endif
 
 /*****************************************************************************/
 /* Internal Programs							     */
@@ -32,20 +35,20 @@
 #else
 #define _ARG(A) ()
 #endif
- int parse _ARG((void));			   /* parse.c                */
+ int parse_bib _ARG((Record rec));		   /* parse.c                */
  int read_rsc _ARG((char *name));		   /* parse.c                */
  int see_bib _ARG((char * fname));		   /* parse.c                */
- int see_bib_msg _ARG((char *s));		   /* parse.c                */
- int see_rsc _ARG((char * fname));		   /* parse.c                */
+ static int see_rsc _ARG((char * fname));	   /* parse.c                */
  int seen _ARG((void));				   /* parse.c                */
  static int fill_line _ARG((void));		   /* parse.c                */
  static int parse_block _ARG((int quotep));	   /* parse.c                */
- static int parse_equation _ARG((int macp));	   /* parse.c                */
+ static int parse_equation _ARG((Record rec,int macp));/* parse.c            */
  static int parse_key _ARG((int alpha));	   /* parse.c                */
  static int parse_rhs _ARG((void));		   /* parse.c                */
  static int parse_string _ARG((int quotep));	   /* parse.c                */
  static int parse_symbol _ARG((int alpha));	   /* parse.c                */
  static int parse_value _ARG((void));		   /* parse.c                */
+ static int see_bib_msg _ARG((char *s));	   /* parse.c                */
  static int skip _ARG((int inc));		   /* parse.c                */
  static int skip_c _ARG((void));		   /* parse.c                */
  static int skip_nl _ARG((void));		   /* parse.c                */
@@ -53,7 +56,6 @@
  static void init_parse _ARG((void));		   /* parse.c                */
  static void parse_number _ARG((void));		   /* parse.c                */
  void init_read _ARG((void));			   /* parse.c                */
- void normalize_symbol _ARG((char * s));	   /* parse.c                */
  void set_rsc_path _ARG((char * val));		   /* parse.c                */
 
 /*****************************************************************************/
@@ -62,23 +64,24 @@
 
 /*---------------------------------------------------------------------------*/
 
-#define Error(X)	    error(ERR_ERROR|ERR_POINT|ERR_FILE,X,empty,empty,\
-				  file_line_buffer,flp,flno,filename)
-#define Warning(X)	    error(ERR_WARN|ERR_POINT|ERR_FILE,X,empty,empty, \
-				  file_line_buffer,flp,flno,filename)
+/*-----------------------------------------------------------------------------
+** Variable*:	parse_sb
+** Purpose:	
+**		
+**		
+**___________________________________________________			     */
+ static StringBuffer * parse_sb = (StringBuffer*)NULL;
 
-
- static StringBuffer *parse_sb = (StringBuffer*)NULL;
- 
 #define FLBLEN	80		/* initial size and increment of line buffer */
 
- char		*filename;
+ static char	*filename;
  static FILE	*file;
- static uchar	*file_line_buffer;
+ static Uchar	*file_line_buffer;
+ static Uchar	*flp;
  static size_t	fl_size	 = 0;
  static int	flno	 = 0;
- static uchar	*flp;
 
+/*---------------------------------------------------------------------------*/
 
 #define EmptyC		(*flp=='\0')
 #define CurrentC	*flp
@@ -94,10 +97,18 @@
 			flp  = file_line_buffer;	\
 			flno = 0;
 
- static char *str_syntax = "Syntax Error";
- static char *str_eof	 = "EOF unexpected";
- static char *str_stdin	 = "<stdin>";
+/*---------------------------------------------------------------------------*/
 
+ static Uchar *str_syntax	= "Syntax Error";
+ static Uchar *str_eof		= "EOF unexpected";
+ static Uchar *str_stdin	= "<stdin>";
+
+#define Error(X)	error(ERR_ERROR|ERR_POINT|ERR_FILE,X,	\
+			      sym_empty,sym_empty,		\
+			      file_line_buffer,flp,flno,filename)
+#define Warning(X)	error(ERR_WARN|ERR_POINT|ERR_FILE,X,	\
+			      sym_empty,sym_empty,		\
+			      file_line_buffer,flp,flno,filename)
 #define SyntaxError	Error(str_syntax)
 #define EofError	Error(str_eof)
 
@@ -120,7 +131,9 @@ static void init___(pathp,pattern,envvp,env)	   /*			     */
 		*cp;				   /*			     */
 						   /*			     */
   if ( *pathp != (char**)0 )			   /*			     */
-  { free((char*)*pathp); *pathp = (char**)0; }	   /*			     */
+  { free((char*)*pathp);			   /*                        */
+    *pathp = (char**)0;				   /*                        */
+  }	   					   /*			     */
 						   /*			     */
   if ( (cp = getenv(env)) != NULL )		   /*			     */
   { *envvp = cp; }				   /*			     */
@@ -151,26 +164,33 @@ static void init___(pathp,pattern,envvp,env)	   /*			     */
 ** Function:	init_read()
 ** Purpose:	Initialize the reading apparatus.
 **		Primarily try to figure out the file search path.
-** Arguments:
-**	
+**
+**		Note that this function is for internal purposes
+**		mainly. The normal user should call |init_bibtool()|
+**		instead. Just in case the search paths are changed
+**		afterwards this function has to be called again to
+**		propagate the information.
+** Arguments:	none
 ** Returns:	nothing
 **___________________________________________________			     */
 void init_read()				   /*			     */
 {						   /*			     */
+#ifndef HAVE_LIBKPATHSEA
   init___(&f_path,				   /*			     */
 	  f_pattern,				   /*			     */
 	  &rsc_v_bibtex,			   /*			     */
 	  rsc_e_bibtex	);			   /*			     */
+#endif
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
 ** Function:	see_bib_msg()
-** Purpose:	Message function for use with  px_fopen().
+** Purpose:	Message function for use with  |px_fopen()|.
 ** Arguments:
-**	s
-** Returns:	
+**	s	String to print
+** Returns:	|TRUE| to indicate that continuation is desired.
 **___________________________________________________			     */
-int see_bib_msg(s)				   /*			     */
+static int see_bib_msg(s)			   /*			     */
   register char *s;				   /*			     */
 {						   /*			     */
   if ( rsc_verbose ) VerbosePrint2("Trying ",s);   /*			     */
@@ -179,44 +199,77 @@ int see_bib_msg(s)				   /*			     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	see_bib()
-** Purpose:	Open a bib file to read from.
-**		If the argument is NULL then stdin is used as input stream.
+** Purpose:	Open a \BibTeX{} file to read from. 
+**		If the argument is |NULL| then |stdin| is used as
+**		input stream.
+**
+**		This function has to be called before |parse()| can be
+**		called. It initializes the parser routine and takes
+**		care that the next reading is done from the given
+**		file.
+**
+**		The file opened with this function has to be closed
+**		with |seen()|.
+**
+**		This function is for internal purposes mainly. See
+**		|read_db()| for a higher level function to read a
+**		database. 
 ** Arguments:
-**	fname
-** Returns:	
+**	fname	Name of the file or |NULL|.
+** Returns:	|TRUE| iff the file couls be opened for reading.
 **___________________________________________________			     */
 int see_bib(fname)				   /*			     */
   register char * fname;			   /*			     */
 {						   /*			     */
   init_parse();					   /*			     */
   InitLine;					   /*			     */
-  if ( fname )					   /*			     */
-  { file = px_fopen(fname,			   /*			     */
-		    "r",			   /*			     */
-		    f_pattern,			   /*			     */
-		    f_path,			   /*			     */
-		    see_bib_msg);		   /*			     */
-    filename = px_filename;			   /*			     */
-    return( file != NULL );			   /*			     */
-  }						   /*			     */
-  filename = str_stdin;				   /*			     */
-  file	   = stdin;				   /*			     */
-  return TRUE;					   /*			     */
+  if ( fname == NULL )				   /*			     */
+  {						   /*                        */
+    filename = str_stdin;			   /*			     */
+    file     = stdin;			   	   /*			     */
+    return TRUE;				   /*			     */
+  }						   /*                        */
+#ifdef HAVE_LIBKPATHSEA
+  filename = kpse_find_file(fname,		   /*                        */
+			    kpse_bib_format,	   /*                        */
+			    TRUE);		   /*                        */
+  if ( filename == NULL ) return FALSE;	   	   /*                        */
+  file = fopen(filename,"r");			   /*                        */
+#else
+  file = px_fopen(fname,			   /*			     */
+		  "r",			   	   /*			     */
+		  f_pattern,			   /*			     */
+		  f_path,			   /*			     */
+		  see_bib_msg);		   	   /*			     */
+  filename = px_filename;			   /*			     */
+#endif
+  return ( file != NULL );			   /*			     */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
 ** Function:	seen()
-** Purpose:	Close input file.
+** Purpose:	Close input file for the \BibTeX{} reading apparatus.
+**		After this function has been called |parse()| might
+**		not return sensible results.
+**
+**		This function is for internal purposes mainly. See
+**		|read_db()| for a higher level function to read a
+**		database. 
 ** Arguments:	none
-** Returns:	Failure status
+** Returns:	|FALSE| if an attempt was made to close an already
+**		closed file.
 **___________________________________________________			     */
 int seen()					   /*			     */
-{ if ( file && file != stdin )			   /*			     */
+{ 						   /*                        */
+  if ( file == stdin )			   	   /*			     */
+  { file = NULL;				   /*                        */
+    return TRUE;				   /*                        */
+  }    						   /*                        */
+  else if ( file )			   	   /*			     */
   { (void)fclose(file);				   /*			     */
-    file = 0L;					   /*			     */
+    file = NULL;				   /*			     */
     return TRUE;				   /*			     */
   }						   /*			     */
-  file = 0L;					   /*			     */
   return FALSE;					   /*			     */
 }						   /*------------------------*/
 
@@ -225,12 +278,13 @@ int seen()					   /*			     */
 #define ExpectSymbol(C,N) if (!parse_symbol(C))		return(N)
 #define ExpectKey(C,N)    if (!parse_key(C))		return(N)
 #define ExpectRhs(N)	  if (!parse_rhs())		return(N)
-#define ExpectEq(N)	  if (!parse_equation(FALSE))	return(N)
-#define ExpectEqMac(N)	  if (!parse_equation(TRUE))	return(N)
+#define ExpectEq(R,N)	  if (!parse_equation(R,FALSE))	return(N)
+#define ExpectEqMac(R,N)  if (!parse_equation(R,TRUE))	return(N)
 
 /*-----------------------------------------------------------------------------
 ** Function:	init_parse()
 ** Purpose:	Initialize the parser.
+**
 ** Arguments:	none
 ** Returns:	nothing
 **___________________________________________________			     */
@@ -239,8 +293,8 @@ static void init_parse()			   /*			     */
   if ( fl_size != 0 ) return;			   /* Already initialzed?    */
 						   /*			     */
   fl_size = FLBLEN;				   /*			     */
-  if ( (file_line_buffer=(uchar*)malloc(fl_size)) == NULL )/* Allocate the   */
-  { OUT_OF_MEMORY("line buffer"); }  		   /*  line buffer	     */
+  if ( (file_line_buffer=(Uchar*)malloc(fl_size*sizeof(Uchar))) == NULL )/*  */
+  { OUT_OF_MEMORY("line buffer"); }		   /* Allocate line buffer   */
 				   		   /*  or message and exit.  */
   if ( parse_sb		   == (StringBuffer*)0 &&  /* Try to initialize the  */
        (parse_sb=sbopen()) == (StringBuffer*)0 )   /*  string buffer for the */
@@ -256,15 +310,15 @@ static void init_parse()			   /*			     */
 **		Since I don't want to use a fixed line length the algorithm
 **		is a little bit more complicated.
 **		First I try to read into the file_line_buffer as it is.
-**		If this succeeds I check to see what caused fgets to terminate.
-**		Either a end-of-line is in the buffer or the buffer is not
-**		filled, i.e. end-of-file has been encountered,
-**		then 0 is returned.
+**		If this succeeds I check to see what caused |fgets| to
+**		terminate. Either a end-of-line is in the buffer or
+**		the buffer is not filled, i.e. end-of-file has been
+**		encountered, then |0| is returned.
 **		Otherwise the buffer has been filled and I try to enlarge the
 **		buffer and read another chunk of bytes.
 ** Arguments:	none
 ** Returns:	Returns 0 iff a character has been read.
-**___________________________________________________			     */
+**___________________________________________________'			     */
 static int fill_line()				   /*			     */
 { register int	len;				   /*			     */
 						   /*			     */
@@ -290,7 +344,7 @@ static int fill_line()				   /*			     */
 	|| len < fl_size-1)			   /*			     */
     { return 0; }				   /*			     */
 						   /*			     */
-    if ( (file_line_buffer = (uchar*)		   /* Try to enlarge	     */
+    if ( (file_line_buffer = (Uchar*)		   /* Try to enlarge	     */
 	  realloc((char*)file_line_buffer,	   /*  the line buffer	     */
 		  fl_size+=FLBLEN)) == NULL )	   /*			     */
     { OUT_OF_MEMORY("line buffer"); }		   /*			     */
@@ -377,20 +431,6 @@ static int skip_nl()				   /*			     */
   }						   /*			     */
 }						   /*------------------------*/
 
-/*-----------------------------------------------------------------------------
-** Function:	normalize_symbol()
-** Purpose:	Function to translate a symbol into a normal form.
-**		This will translate the symbol to lower case.
-** Arguments:
-**	s
-** Returns:	nothing
-**___________________________________________________			     */
-void normalize_symbol(s)			   /*                        */
-  register char * s;				   /*                        */
-{ 						   /*                        */
-  while ( *s ) { *s = ToLower(*s); ++s; }	   /*                        */
-}						   /*------------------------*/
-
 /*-----------------------------------------------------------------------------
 ** Function:	parse_symbol()
 ** Purpose:	Parse a symbol and push it to the stack.
@@ -404,13 +444,13 @@ static int parse_symbol(alpha)			   /*			     */
 { register int c;				   /*			     */
   register char * cp;				   /*			     */
 						   /*			     */
-  c = GetC;  cp = (char*)flp-1;			   /*			     */
+  c = GetC;					   /*                        */
+  cp = (char*)flp-1;			   	   /*			     */
   if ( alpha && (! is_alpha(c)) )		   /*			     */
   { SyntaxError; return(FALSE); }		   /*			     */
   while ( is_allowed(CurrentC) ) { SkipC; }	   /*			     */
   c = CurrentC; CurrentC = '\0';		   /*			     */
-  normalize_symbol(cp);				   /*			     */
-  push_string(symbol(cp));			   /*			     */
+  push_string(symbol(lower(cp)));		   /*			     */
   CurrentC = c;					   /*			     */
   return TRUE;					   /*			     */
 }						   /*------------------------*/
@@ -435,13 +475,11 @@ static int parse_key(alpha)			   /*			     */
   c = CurrentC; CurrentC = '\0';		   /*			     */
   if ( rsc_key_case )				   /*                        */
   { s = symbol(cp);				   /*                        */
-    normalize_symbol(cp);			   /*			     */
-    cp = symbol(cp);				   /*                        */
+    cp = symbol(lower(cp));			   /*                        */
     save_key(cp,s);				   /*                        */
   }						   /*                        */
   else						   /*                        */
-  { normalize_symbol(cp);			   /*			     */
-    cp = symbol(cp);				   /*                        */
+  { cp = symbol(lower(cp));			   /*                        */
   }						   /*                        */
   push_string(cp);			   	   /*			     */
   CurrentC = c;					   /*			     */
@@ -476,14 +514,14 @@ static void parse_number()			   /*			     */
 **	quotep	Boolean. TRUE iff the leading " has already been stripped off
 **		the input stream.
 ** Returns:	Success status
-**___________________________________________________			     */
+**___________________________________________________"			     */
 static int parse_string(quotep)			   /*			     */
   int          quotep;				   /*			     */
 { register int c,				   /*                        */
     	       left;				   /*			     */
 						   /*			     */
   left = 0;					   /*			     */
-  if ( quotep ) (void)sbputchar('"',parse_sb);	   /*			     */
+  if ( quotep ) (void)sbputchar('"',parse_sb);	   /*"			     */
   do						   /*			     */
   { switch ( c=skip_nl() )			   /*			     */
     { case EOF:	 EofError; return(FALSE);	   /*			     */
@@ -515,7 +553,7 @@ static int parse_string(quotep)			   /*			     */
 **___________________________________________________			     */
 static int parse_block(quotep)			   /*			     */
   int quotep;					   /*			     */
-{ register int c,left;				   /*			     */
+{ register int c, left;				   /*			     */
 						   /*			     */
   left = 1;					   /*			     */
   if ( quotep ) (void)sbputchar('{',parse_sb);	   /*			     */
@@ -568,7 +606,9 @@ static int parse_rhs()				   /*			     */
 	UnGetC; ExpectSymbol(TRUE,FALSE);	   /*			     */
 	{ register char * mac;			   /*			     */
 	  mac = pop_string();			   /*			     */
+#ifdef OLD
 	  (void)look_macro(mac,1);		   /*			     */
+#endif
 	  (void)sbputs(mac,parse_sb);		   /*			     */
 	}					   /*			     */
     }						   /*			     */
@@ -585,11 +625,13 @@ static int parse_rhs()				   /*			     */
 **		
 **
 ** Arguments:
+**	rec
 **	macp
 ** Returns:	Success status
 **___________________________________________________			     */
-static int parse_equation(macp)			   /*			     */
-  int macp;					   /*			     */
+static int parse_equation(rec,macp)		   /*			     */
+  Record rec;					   /*                        */
+  int    macp;					   /*			     */
 { register char *s, *t;				   /*			     */
 						   /*			     */
   ExpectSymbol(TRUE,FALSE);			   /*			     */
@@ -598,105 +640,158 @@ static int parse_equation(macp)			   /*			     */
 						   /*			     */
   t = pop_string();				   /*			     */
   s = pop_string();				   /*			     */
-  push_to_master_record(s,t);			   /*			     */
+  push_to_record(rec,s,t);			   /*			     */
+#ifdef OLD
   if ( macp ) { def_macro(s,t,0); }		   /*			     */
+#endif
   return(TRUE);					   /*			     */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:	parse()
+** Function:	parse_bib()
 ** Purpose:	Read one entry and fill the internal record structure.
-**		return the type of the entry.
-** Arguments:	none
-** Returns:	
+**		Return the type of the entry read.
+**
+**		|BIB_EOF| is returned if nothing could be read and
+**		the end of the file has been encountered.
+**
+**		|BIB_NOOP| is returned when an error has occured. This is
+**		an indicator that no record has been read but the
+**		error recovery is ready to try it again.
+**
+**		This function is for internal purposes mainly. See
+**		|read_db()| for a higher level function to read a
+**		database. 
+** Arguments:
+**	rec	Record to store the result in.
+** Returns:	The type of the entry read, |BIB_EOF|, or |BIB_NOOP|.
 **___________________________________________________			     */
-int parse()					   /*			     */
+int parse_bib(rec)				   /*			     */
+  Record rec;					   /*                        */
 { register int type,				   /*			     */
 	       n,				   /*			     */
 	       c,				   /*			     */
 	       om;				   /*			     */
   long	       ignored = 0L;			   /*			     */
+  char         *name;				   /*                        */
+  int          line;				   /*                        */
+  char	       buffer[32];			   /*                        */
+  static StringBuffer * comment_sb = (StringBuffer*)NULL;/*                  */
+ 						   /*                        */
+  if ( file == NULL ) return BIB_EOF;	   	   /*                        */
+  if ( comment_sb == (StringBuffer*)NULL )	   /*                        */
+  { comment_sb = sbopen(); } 			   /*                        */
+ 						   /*                        */
+  RecordOldKey(rec)	 = NULL;	   	   /*			     */
+  RecordFree(rec)	 = 0;		   	   /*			     */
+  RecordComment(rec)	 = sym_empty;	   	   /*                        */
  						   /*                        */
   do						   /*                        */
   { init_parse();				   /*			     */
 						   /*			     */
     while ( (c=skip_c()) != '@' )		   /* Skip to next @	     */
-    { if ( c==EOF ) return(EndOfFile);		   /*			     */
+    { if ( c==EOF )				   /*                        */
+      { char *s, *t;				   /*                        */
+	if ( ignored == 0 ) return(BIB_EOF); 	   /*			     */
+	RecordType(rec) = BIB_COMMENT;		   /*                        */
+ 						   /*                        */
+	s = t = sbflush(comment_sb);
+	while ( *s ) s++;
+	while ( t <= --s  && isspace(*s) ) *s = '\0';
+	if ( *t ) RecordComment(rec) = symbol(t);
+	sbrewind(comment_sb);			   /*                        */
+	return(BIB_COMMENT);		   	   /*			     */
+      }						   /*                        */
       if ( !is_space(c) ) ++ignored;		   /*			     */
       if ( ignored > 0 && rsc_pass_comment )	   /*			     */
-      { putc_in(c); }				   /*			     */
+      { sbputchar(c,comment_sb); }		   /*			     */
     }						   /*			     */
     						   /*			     */
     if ( ignored != 0L )			   /*			     */
-    { if ( rsc_pass_comment ) putc_in('\n');	   /*			     */
+    { if ( rsc_pass_comment )			   /*                        */
+      { sbputchar('\n',comment_sb); }	   	   /*			     */
       else					   /*			     */
-      { static char buffer[32];			   /*			     */
-      						   /*			     */
-	(void)sprintf(buffer,"%ld",ignored);	   /*			     */
+      { (void)sprintf(buffer,"%ld",ignored);	   /*			     */
 	error(ERR_WARN|ERR_FILE,buffer,		   /*			     */
 	      " non-space characters ignored.",	   /*			     */
-	      (char*)0,(uchar*)0,(uchar*)0,	   /*			     */
+	      (char*)0,(Uchar*)0,(Uchar*)0,	   /*			     */
 	      flno,filename);			   /*			     */
       }						   /*			     */
     }						   /*			     */
 						   /*			     */
-    if ( (type=find_entry_type((char*)flp)) == NOOP )/*			     */
-    { Error("Unknown entry type"); return(NOOP); } /*			     */
+    if ( (type=find_entry_type((char*)flp)) == BIB_NOOP )/*		     */
+    { Error("Unknown entry type"); return(BIB_NOOP); }/*		     */
     flp += strlen(EntryName(type));		   /*			     */
  						   /*                        */
-    if ( type == COMMENT && rsc_pass_comment )
-    { putc_in('@');
-      puts_in(EntryName(type),0);
-    }
-  } while (type == COMMENT);			   /*                        */
+    if ( type == BIB_COMMENT && rsc_pass_comment ) /*                        */
+    { sbputchar('@',comment_sb);		   /*                        */
+      sbputs(EntryName(type),comment_sb);	   /*                        */
+    }						   /*                        */
+  } while (type == BIB_COMMENT);		   /*                        */
 					   	   /*			     */
   c = GetC;					   /*			     */
   if ( c != '{' && c != '(' )			   /*			     */
   { Error("Expected '{' or '(' missing");	   /*			     */
-    return(NOOP);				   /*			     */
+    return(BIB_NOOP);				   /*			     */
   }						   /*			     */
+  line = flno;					   /*                        */
+  name = filename;				   /*                        */
 						   /*			     */
-  clear_master_record(type);			   /*			     */
+  RecordType(rec) = type;			   /*                        */
 						   /*			     */
   switch ( type )				   /*			     */
-  { case COMMENT:				   /* This code is not used  */
+  { case BIB_COMMENT:				   /* This code is not used  */
       UnGetC; 					   /*  any more.             */
       (void)parse_rhs();			   /*                        */
-      push_to_master_record(pop_string(),(char*)0);/*			     */
+      push_to_record(rec,pop_string(),(char*)0);   /*			     */
       return type;				   /*                        */
  						   /*                        */
-    case PREAMBLE:				   /*			     */
-      ExpectRhs(NOOP);				   /*			     */
-      push_to_master_record(pop_string(),(char*)0);/*			     */
+    case BIB_PREAMBLE:				   /*			     */
+      ExpectRhs(BIB_NOOP);			   /*			     */
+      push_to_record(rec,pop_string(),(char*)0);   /*			     */
       break;					   /*			     */
 						   /*			     */
-    case STRING:				   /*			     */
-      ExpectEqMac(NOOP);			   /*			     */
+    case BIB_STRING:				   /*			     */
+      ExpectEqMac(rec,BIB_NOOP);		   /*			     */
       break;					   /*			     */
 						   /*			     */
+    case BIB_ALIAS:				   /*			     */
+      ExpectEq(rec,BIB_NOOP);			   /*			     */
+      break;					   /*			     */
+						   /*			     */
+    case BIB_INCLUDE:				   /*			     */
+      ExpectRhs(BIB_NOOP);			   /*			     */
+      push_to_record(rec,pop_string(),(char*)0);   /*			     */
+      break;					   /*			     */
+						   /*			     */
+    case BIB_MODIFY:				   /*			     */
     default:					   /*			     */
       if ( TestC == ',' )			   /*			     */
       { Warning("Missing reference key");	   /*			     */
-	push_to_master_record("",(char*)0);	   /*			     */
+	push_to_record(rec,sym_empty,(char*)0);	   /*			     */
 	(void)GetC;				   /*			     */
       }						   /*			     */
       else					   /*			     */
-      { ExpectKey(FALSE,NOOP);		   	   /*			     */
-	Expect(',',NOOP);			   /*			     */
-	push_to_master_record(pop_string(),(char*)0);/*			     */
+      { ExpectKey(FALSE,BIB_NOOP);		   /*			     */
+	Expect(',',BIB_NOOP);			   /*			     */
+	push_to_record(rec,pop_string(),(char*)0); /*			     */
       }						   /*			     */
 						   /*			     */
       do					   /*			     */
-      { ExpectEq(NOOP);				   /*			     */
+      { ExpectEq(rec,BIB_NOOP);			   /*			     */
 	for( n=0; GetC==','; n++)		   /*			     */
 	{ if ( n > 0 )				   /*			     */
 	  { Warning("Multiple ',' ignored."); }	   /*			     */
 	}					   /*			     */
 	UnGetC;					   /*			     */
-	om = ( (TestC != '}') && (TestC != ')') ); /*			     */
-	if ( om && n==0 )			   /*			     */
-	{ Warning("Missing ',' assumed."); }	   /*			     */
-						   /*			     */
+	switch (TestC)				   /*                        */
+	{ case EOF:				   /*                        */
+	  case '}':				   /*                        */
+	  case ')': om = FALSE; break;		   /*                        */
+	  default:  om = TRUE;			   /*                        */
+	    if ( n==0 )			   	   /*			     */
+	    { Warning("Missing ',' assumed."); }   /*			     */
+	}					   /*                        */
       } while ( om );				   /*			     */
   }						   /*			     */
 						   /*			     */
@@ -709,10 +804,33 @@ int parse()					   /*			     */
       if( c != '(' )				   /*			     */
       { Warning("Parenthesis '{' closed by ')'"); }/*			     */
       break;					   /*			     */
+    case EOF:					   /*                        */
+      { char *s;				   /*                        */
+        if (c=='{') { s="'{'"; } else { s="'('"; } /*                        */
+	error(ERR_ERROR|ERR_FILE,s,		   /*                        */
+	      " not closed at end of file.",	   /*                        */
+	      (char*)0,(Uchar*)0,(Uchar*)0,	   /*			     */
+	      line,name);			   /*			     */
+      }						   /*                        */
+      break;					   /*                        */
     default:					   /*			     */
-      Error("Expected '}' or ')' missing");	   /*			     */
-      return(NOOP);				   /*			     */
+      { char *s;				   /*                        */
+        if (c=='{') { s="'{'"; } else { s="'('"; } /*                        */
+	error(ERR_ERROR|ERR_FILE,		   /*                        */
+	      s," not properly terminated.", 	   /*                        */
+	      (char*)0,(Uchar*)0,(Uchar*)0, 	   /*			     */
+	      line,name);			   /*			     */
+      }						   /*                        */
+      return(BIB_NOOP);				   /*			     */
   }						   /*			     */
+ 						   /*                        */
+  { char *s, *t;				   /*                        */
+    s = t = sbflush(comment_sb);		   /*                        */
+    while ( *s ) s++;				   /*                        */
+    while ( t <= --s  && isspace(*s) ) *s = '\0';  /*                        */
+    if ( *t ) RecordComment(rec) = symbol(t);	   /*                        */
+  }						   /*                        */
+  sbrewind(comment_sb);				   /*                        */
   return(type);					   /*			     */
 }						   /*------------------------*/
 
@@ -730,11 +848,11 @@ int parse()					   /*			     */
 ** Purpose:	Initialize the resource file reading apparatus.
 **		Primarily try to figure out the file search path.
 ** Arguments:
-**	val
+**	val	The string representation of the file search path.
 ** Returns:	nothing
 **___________________________________________________			     */
 void set_rsc_path(val)				   /*			     */
-  register char * val;				   /*			     */
+  char * val;				   	   /*			     */
 {						   /*			     */
   rsc_v_rsc = val;				   /*			     */
   init___(&r_path,				   /*			     */
@@ -745,13 +863,15 @@ void set_rsc_path(val)				   /*			     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	see_rsc()
-** Purpose:	Open a rsc file to read from.
+** Purpose:	Open a rsc file to read from. The resource file path
+**		and the optional extensions are used to construct the
+**		full file name.
 ** Arguments:
-**	fname
-** Returns:	
+**	fname	The file name to take into account.
+** Returns:	|TRUE| iff the operation succeeds.
 **___________________________________________________			     */
-int see_rsc(fname)				   /*			     */
-  register char * fname;			   /*			     */
+static int see_rsc(fname)			   /*			     */
+  char * fname;			   		   /*			     */
 {						   /*			     */
   if ( fname )					   /*			     */
   { init_parse();				   /*			     */
@@ -808,11 +928,19 @@ static int parse_value()			   /*			     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	read_rsc()
-** Purpose:	
-**		
+** Purpose:	Read a resource file and evaluate all instructions
+**		contained.
 **
+**		The characters |#|, |%|, and |;| start an endline
+**		comment but only between resource instructions. They
+**		are not recogniized between a resource instructiuon
+**		and its value or inside the value braces.
+**
+**		This function is contained in this module because it
+**		shares several functions with the \BibTeX{} parsing
+**		routines. 
 ** Arguments:
-**	name
+**	name	Name of the file to read from.
 ** Returns:	
 **___________________________________________________			     */
 int read_rsc(name)				   /*			     */
@@ -822,11 +950,13 @@ int read_rsc(name)				   /*			     */
  						   /*                        */
   char		*s_filename;			   /*			     */
   FILE		*s_file;			   /*                        */
-  uchar		*s_file_line_buffer;		   /*                        */
+  Uchar		*s_file_line_buffer;		   /*                        */
   size_t	s_fl_size;			   /*                        */
   int		s_flno;				   /*                        */
-  uchar		*s_flp;				   /*                        */
- 						   /*                        */
+  Uchar		*s_flp;				   /*                        */
+ 						   /* Save the old state in  */
+ 						   /*  local variables to    */
+ 						   /*  restore it later.     */
   s_filename	     = filename;		   /*			     */
   s_file	     = file;			   /*                        */
   s_file_line_buffer = file_line_buffer;	   /*                        */
@@ -855,10 +985,10 @@ int read_rsc(name)				   /*			     */
 	}					   /*			     */
     }						   /*			     */
     (void)seen();				   /*			     */
-    c = 0;					   /*			     */
+    c = TRUE;					   /*			     */
   }						   /*                        */
   else						   /*                        */
-  { c = 1;		   			   /*			     */
+  { c = FALSE;		   			   /*			     */
   }						   /*                        */
   if ( fl_size > 0 ) 				   /*                        */
   { free((char*)file_line_buffer);		   /*                        */
