@@ -1,26 +1,33 @@
 /******************************************************************************
-** $Id: print.c,v 2.23 1996/02/22 21:43:51 gerd Exp gerd $
+** $Id: print.c,v 2.30 1997/12/06 18:57:08 gerd Exp $
 **=============================================================================
 ** 
 ** This file is part of BibTool.
 ** It is distributed under the GNU General Public License.
 ** See the file COPYING for details.
 ** 
-** (c) 1996 Gerd Neugebauer
+** (c) 1996-1997 Gerd Neugebauer
 ** 
 ** Net: gerd@informatik.uni-koblenz.de
 ** 
+**-----------------------------------------------------------------------------
+** Description:
+**	This module provides also access to the functions and
+**	variables defined in |entry.c|. Consult also the documentation
+**	of this file for details.
+**
+**
 ******************************************************************************/
 
-#include "bibtool.h"
-#include "print.h"
-#include "entry.h"
-#include "rsc.h"
-#include "macros.h"
-#include "type.h"
-#include "sbuffer.h"
-#include "expand.h"
-#include "error.h"
+#include <bibtool/general.h>
+#include <bibtool/symbols.h>
+#include <bibtool/print.h>
+#include <bibtool/entry.h>
+#include <bibtool/rsc.h>
+#include <bibtool/macros.h>
+#include <bibtool/sbuffer.h>
+#include <bibtool/expand.h>
+#include <bibtool/error.h>
 
 /*****************************************************************************/
 /* Internal Programs							     */
@@ -31,52 +38,55 @@
 #else
 #define _ARG(A) ()
 #endif
- int tell _ARG((char * fname));			   /* print.c                */
- int told _ARG((void));				   /* print.c                */
- static void indent _ARG((int col));		   /* print.c                */
- static void line_breaking _ARG((char *t,int align));/* print.c              */
- static void pretty_print _ARG((Record rec,char *at));/* print.c             */
- static void print_equation _ARG((char *s,char *t,int align));/* print.c     */
- void puts_in _ARG((char *s, int in));	   	   /* print.c                */
- void print_commented _ARG((Record rec));	   /* print.c                */
- void print_record _ARG((Record rec));		   /* print.c                */
- void putc_in _ARG((int c));			   /* print.c                */
+ char * sput_record _ARG((Record rec,DB db,char *start));/* print.c          */
+ static int fput_char _ARG((int c));		   /* print.c                */
+ static int sput_char _ARG((int c));		   /* print.c                */
+ static void indent _ARG((int col,int (*fct)_ARG((int))));/* print.c         */
+ static void line_breaking _ARG((char *t,int align,int (*fct)_ARG((int))));/* print.c*/
+ static void print_equation _ARG((char *pre,char *s,char *t,int align,int (*fct)_ARG((int))));/* print.c*/
+ static void puts_in _ARG((char *s,int in,int (*fct)_ARG((int))));/* print.c */
+ void fput_record _ARG((FILE *file,Record rec,DB db,char *start));/* print.c */
+ void put_record _ARG((int (*fct)_ARG((int)),Record rec,DB db,char *start));/* print.c*/
+ void set_key_type _ARG((char * s));		   /* print.c                */
  void set_symbol_type _ARG((char * s));		   /* print.c                */
 
 /*****************************************************************************/
 /* External Programs							     */
 /*===========================================================================*/
 
- extern int rsc_braces;
-
 /*---------------------------------------------------------------------------*/
 
 #define TAB_WIDTH 8
 
- static FILE * ofile = stdout;
-
-
  static int symbol_type = SYMBOL_TYPE_LOWER;
 
- static char * upper = "upper";
- static char * lower = "lower";
- static char * cased = "cased";
+ static char * s_upper = "upper";
+ static char * s_lower = "lower";
+ static char * s_cased = "cased";
 
 /*-----------------------------------------------------------------------------
 ** Function:	set_symbol_type()
-** Purpose:	Wrapper function to set the static variable symbol_type.
-**		This function is called from rsc.c
+** Purpose:	Function to set the symbol type which is used by the
+**		printing routine. The argument is a string describing
+**		the value to use. Possible values are |"upper"|,
+**		|"lower"|, and |"cased"|. The comparison of the values
+**		is performed case insensitive.
+**
+**		If no appropriate value is found then an error message
+**		is issued as the only action.
+**		
+**		This function is called from |rsc.c|.
 ** Arguments:
 **	s	String description of the value.
 ** Returns:	nothing
 **___________________________________________________			     */
 void set_symbol_type(s)				   /*			     */
   register char * s;				   /*			     */
-{ if ( case_cmp(s,upper) )			   /*			     */
+{ if ( case_cmp(s,s_upper) )			   /*			     */
   { symbol_type = SYMBOL_TYPE_UPPER; }		   /*			     */
-  else if ( case_cmp(s,cased) )		   	   /*			     */
+  else if ( case_cmp(s,s_cased) )		   /*			     */
   { symbol_type = SYMBOL_TYPE_CASED; }		   /*			     */
-  else if ( case_cmp(s,lower) )		   	   /*			     */
+  else if ( case_cmp(s,s_lower) )		   /*			     */
   { symbol_type = SYMBOL_TYPE_LOWER; }		   /*			     */
   else						   /*			     */
   { Err("Unknown symbol type ignored.\n"); }	   /*			     */
@@ -88,7 +98,7 @@ void set_symbol_type(s)				   /*			     */
  static int key_type = SYMBOL_TYPE_LOWER;
 
 /*-----------------------------------------------------------------------------
-** Function:	set_key_type()
+** Function*:	set_key_type()
 ** Purpose:	Wrapper function to set the static variable key_type.
 **		This function is called from rsc.c
 ** Arguments:
@@ -97,65 +107,25 @@ void set_symbol_type(s)				   /*			     */
 **___________________________________________________			     */
 void set_key_type(s)				   /*			     */
   register char * s;				   /*			     */
-{ if ( case_cmp(s,upper) )			   /*			     */
+{ if ( case_cmp(s,s_upper) )			   /*			     */
   { key_type = SYMBOL_TYPE_UPPER; }		   /*			     */
-  else if ( case_cmp(s,cased) )		   	   /*			     */
+  else if ( case_cmp(s,s_cased) )		   /*			     */
   { key_type = SYMBOL_TYPE_CASED; }		   /*			     */
-  else if ( case_cmp(s,lower) )		   	   /*			     */
+  else if ( case_cmp(s,s_lower) )		   /*			     */
   { key_type = SYMBOL_TYPE_LOWER; }		   /*			     */
   else						   /*			     */
   { Err("Unknown key type ignored.\n"); }	   /*			     */
 }						   /*------------------------*/
-
 #endif
-
-
-/*-----------------------------------------------------------------------------
-** Function:	tell()
-** Purpose:	Redirect further output to the file given.
-**		To avoid problems with a library function of this name,
-**		the name is remapped in a header file.
-**		The tell()/told() is inpired by the Prolog predicates of
-**		the same names.
-** Arguments:
-**	fname	File name of the new output stream.
-** Returns:	
-**___________________________________________________			     */
-int tell(fname)					   /*			     */
-  register char * fname;			   /*			     */
-{						   /*			     */
-  if ( ofile != stdout ) (void)fclose(ofile);	   /*			     */
-  if ( fname == NULL ||				   /*			     */
-       (ofile = fopen(fname,"w"))==NULL )	   /*			     */
-  { ofile = stdout; return FALSE; }		   /*			     */
-  return TRUE;					   /*			     */
-}						   /*------------------------*/
-
-/*-----------------------------------------------------------------------------
-** Function:	told()
-** Purpose:	Close formerly opened output file.
-**		Reset the output file to stdout.
-** Arguments:	none
-** Returns:	TRUE
-**___________________________________________________			     */
-int told()					   /*			     */
-{						   /*			     */
-  if ( ofile != stdout )			   /*			     */
-  { (void)fclose(ofile);			   /*			     */
-    ofile = stdout;				   /*			     */
-  }						   /*			     */
-  return TRUE;					   /*			     */
-}						   /*------------------------*/
-
 
 						   /*------------------------*/
  static int column = 0;				   /* The current column of  */
 						   /*  the output stream is  */
 						   /*  kept in this variable.*/
 
-#define NL	(void)fputc('\n',ofile),column=0
-#define PUTC(C) (void)(fputc(C,ofile),++column)
-#define PUTS(S) puts_in(S,0)
+#define NL	(void)(*fct)('\n'),column=0
+#define PUTC(C) (void)((*fct)(C),++column)
+#define PUTS(S) puts_in(S,0,fct)
 
 /*-----------------------------------------------------------------------------
 ** Function:	puts_in()
@@ -163,51 +133,29 @@ int told()					   /*			     */
 ** Arguments:
 **	s	string to be printed.
 **	in	indentation. Alignment column.
+**	fct	function to use for writing a character.
 ** Returns:	nothing
 **___________________________________________________			     */
-void puts_in(s,in)			   	   /*			     */
+static void puts_in(s,in,fct)			   /*			     */
   register char *s;				   /*			     */
   register int  in;				   /*                        */
+  int (*fct)_ARG((int));			   /*                        */
 {						   /*			     */
   while ( *s )					   /*			     */
-  { (void)putc(*s,ofile);			   /*                        */
+  { (void)(*fct)(*s);			   	   /*                        */
     switch ( *(s++) )				   /*			     */
     { case '\t':				   /*                        */
 	column += TAB_WIDTH - (column%TAB_WIDTH);  /*                        */
 	break;					   /*	                     */
       case '\n':				   /*                        */
 	column = 0;			   	   /*                        */
-        if ( in > 0 ) indent(in);		   /*                        */
+        if ( in > 0 ) indent(in,fct);		   /*                        */
         break;  				   /*			     */
       default:	 ++column;			   /*			     */
     }						   /*			     */
   }						   /*			     */
 }						   /*------------------------*/
 
-/*-----------------------------------------------------------------------------
-** Function:	putc_in()
-** Purpose:	Print a character and update the current column.
-** Arguments:
-**	c	character to be printed.
-** Returns:	nothing
-**___________________________________________________			     */
-void putc_in(c)					   /*			     */
-  register int c;				   /*			     */
-{						   /*			     */
-  (void)putc(c,ofile);				   /*			     */
- 						   /*                        */
-  switch ( c )					   /*			     */
-  { case '\t':					   /*                        */
-      column += TAB_WIDTH - (column%TAB_WIDTH);	   /*                        */
-      break;					   /*	                     */
-    case '\n':					   /*                        */
-      column = 0;				   /*                        */
-      break;	   				   /*			     */
-    default:					   /*                        */
-      ++column;			   		   /*			     */
-  }						   /*			     */
-}						   /*------------------------*/
-
 /*-----------------------------------------------------------------------------
 ** Function:	indent()
 ** Purpose:	Add spaces or tabs to indent to the given column.
@@ -215,19 +163,22 @@ void putc_in(c)					   /*			     */
 **		The resource use.tabs can be used to disable the use of TAB.
 ** Arguments:
 **	col	Target column
+**	fct	function to use for writing a character.
 ** Returns:	nothing
 **___________________________________________________			     */
-static void indent(col)				   /*			     */
+static void indent(col,fct)			   /*			     */
   register int col;				   /*			     */
+  int (*fct)_ARG((int));			   /*                        */
 {						   /*			     */
+  if ( col > rsc_linelen ) col = rsc_linelen;	   /*                        */
   while ( column < col )			   /*			     */
   { if (   rsc_use_tabs				   /*	TAB is allowed and   */
 	&& column+TAB_WIDTH-(column%TAB_WIDTH) <= col )/* enough space left  */
-    { (void)putc('\t',ofile);			   /*	then put a TAB and   */
+    { (void)(*fct)('\t');			   /*	then put a TAB and   */
       column += TAB_WIDTH - (column%TAB_WIDTH);	   /*	update column.	     */
     }						   /*			     */
     else					   /* otherwise		     */
-    { (void)putc(' ',ofile);			   /*  write a singe space   */
+    { (void)(*fct)(' ');			   /*  write a singe space   */
       ++column;					   /*  and advance column.   */
     }						   /*			     */
   }						   /*			     */
@@ -236,18 +187,20 @@ static void indent(col)				   /*			     */
 /*-----------------------------------------------------------------------------
 ** Function:	line_breaking()
 ** Purpose:	Write out a right hand side of an equation.
-**		If it doesn't fit break the line into several parts and
+**		If it does not fit break the line into several parts and
 **		print them on successive lines.
 **		Temporarily end marks are placed inside the string,
 **		but the old contents has been restored at the end.
 ** Arguments:
 **	t	string to print.
 **	align	starting column for continuation lines.
+**	fct	function to use for writing a character.
 ** Returns:	nothing
 **___________________________________________________			     */
-static void line_breaking(t,align)		   /*			     */
+static void line_breaking(t,align,fct)		   /*			     */
   register char *t;				   /* string to print.	     */
   int		align;				   /* alignment column	     */
+  int (*fct)_ARG((int));			   /*                        */
 { register char *s;				   /* intermediate pointer   */
   char		end_c;				   /* temp. character.	     */
   int		brace,				   /* brace counter	     */
@@ -256,7 +209,7 @@ static void line_breaking(t,align)		   /*			     */
 						   /*			     */
   while ( is_space(*t) ) ++t;			   /* skip leading spaces    */
 						   /*			     */
-  indent(align);				   /* goto alignment column  */
+  indent(align,fct);				   /* goto alignment column  */
 						   /*			     */
   while ( *t )					   /* as long as sth to print*/
   { s = t;					   /*			     */
@@ -301,11 +254,11 @@ static void line_breaking(t,align)		   /*			     */
     while ( *s )				   /*			     */
     { if ( len + (first?0:3) <= rsc_linelen - column)/* Is there enough space*/
       { if ( !first ) PUTS(" # ");		   /* Maybe add separator    */
-	puts_in(s,align);			   /* write it out	     */
+	puts_in(s,align,fct);			   /* write it out	     */
 	s = t;					   /* and we are done	     */
       }						   /*			     */
       else if ( !first )			   /* If sth has been before */
-      { puts_in("\n# ",align-2);	   	   /*  start a new line	     */
+      { puts_in("\n# ",align-2,fct);	   	   /*  start a new line	     */
 	first = TRUE;				   /*			     */
       }						   /* Now we have to break   */
       else					   /*  a single entry	     */
@@ -313,7 +266,11 @@ static void line_breaking(t,align)		   /*			     */
 	     *save_ptr,				   /*                        */
 	     *ptr;			   	   /*			     */
 						   /*			     */
-	save_ptr = s + rsc_linelen - column;	   /* Potential end	     */
+        if ( 0 <= rsc_linelen - column )	   /*                        */
+	  save_ptr = s + rsc_linelen - column;	   /* Potential end	     */
+	else					   /*                        */
+	  save_ptr = s;				   /*                        */
+ 						   /*                        */
 	for(ptr=s;				   /* Search next newline    */
 	    ptr <= save_ptr && *ptr != '\n';	   /*  or end of region      */
 	    ptr++) {}				   /*                        */
@@ -321,7 +278,9 @@ static void line_breaking(t,align)		   /*			     */
 	if ( *ptr == '\n' )			   /*                        */
 	{ save_ptr = ptr;			   /*                        */
 	  *save_ptr = '\0';	   		   /* Save and mark end.     */
-	  puts_in(s,align); NL; indent(align);	   /*			     */
+	  puts_in(s,align,fct);			   /*                        */
+	  NL;					   /*                        */
+	  indent(align,fct);	   		   /*			     */
 	  *save_ptr = '\n';			   /* Restore end	     */
 	  len += s - save_ptr - 1;		   /* Update the length	     */
 	  s = save_ptr+1;			   /*			     */
@@ -337,7 +296,9 @@ static void line_breaking(t,align)		   /*			     */
 	  }					   /*                        */
 	  len += s - save_ptr;			   /* Update the length	     */
 	  save_c = *save_ptr; *save_ptr = '\0';	   /* Save and mark end.     */
-	  puts_in(s,align); NL; indent(align);	   /*			     */
+	  puts_in(s,align,fct);			   /*                        */
+	  NL;					   /*                        */
+	  indent(align,fct);	   		   /*			     */
 	  *save_ptr = save_c;			   /* Restore end	     */
 	  s = save_ptr;	   			   /*			     */
 	  while ( is_space(*s) ) { ++s; len--; }   /* Skip spaces	     */
@@ -359,77 +320,189 @@ static void line_breaking(t,align)		   /*			     */
 **		If desired it can be indented. t is broken if it doesn't fit
 **		in one line.
 ** Arguments:
+**	pre	string to be printed before s.
 **	s	left hand side
 **	t	right hand side
 **	align	target column. If negative no indentation is performed.
+**	fct	function to use for writing a character.
 ** Returns:	nothing
-**___________________________________________________			     */
-static void print_equation(s,t,align)		   /*			     */
-  register char *s;				   /*			     */
-  register char *t;				   /*			     */
-  int		align;				   /*			     */
+**___________________________________________________'			     */
+static void print_equation(pre,s,t,align,fct)	   /*			     */
+  char *pre;					   /*                        */
+  char *s;				   	   /*			     */
+  char *t;				   	   /*			     */
+  int  align;				   	   /*			     */
+  int (*fct)_ARG((int));			   /*                        */
 {						   /*			     */
   if ( align < 0 )				   /*			     */
-  { PUTS(get_item(s,symbol_type));		   /*			     */
-    PUTC('=');					   /*			     */
+  { PUTS(pre);		   			   /*			     */
+    PUTS(get_item(s,symbol_type));		   /*			     */
+    if ( rsc_print_we )				   /*                        */
+    { PUTS(" = "); }				   /*                        */
+    else					   /*                        */
+    { PUTC('='); }				   /*                        */
     PUTS(t);					   /*			     */
   }						   /*			     */
   else						   /*			     */
-  { indent(rsc_indent);				   /*			     */
+  { indent(rsc_indent,fct);			   /*			     */
+    PUTS(pre);		   			   /*			     */
     PUTS(get_item(s,symbol_type));		   /*			     */
-    indent(align-2);				   /*			     */
+    if ( column>=align-2 && rsc_print_we )	   /*                        */
+    { PUTC(' '); }				   /*                        */
+    else if ( rsc_eq_right )			   /*                        */
+    { indent(align-2,fct); }			   /*			     */
+    else if ( column<align || rsc_print_we )	   /*                        */
+    { PUTC(' '); }	   			   /*                        */
     PUTC('=');					   /*			     */
-    line_breaking(t,align);			   /*			     */
+    if ( rsc_print_we )	{ PUTC(' '); }		   /*                        */
+    line_breaking(t,align,fct);			   /*			     */
   }						   /*			     */
 }						   /*------------------------*/
 
+ static FILE * ofile = stdout;
+
 /*-----------------------------------------------------------------------------
-** Function:	print_record()
-** Purpose:	Print a record to the assigned output file
+** Function:	fput_char()
+** Purpose:	Output function which places the character on the |ofile| 
+**		stream.
 ** Arguments:
-**	rec	record to print
-** Returns:	nothing
+**	c	Character to print.
+** Returns:	The return status of |fputc()|.
 **___________________________________________________			     */
-void print_record(rec)				   /*			     */
-  register Record rec;				   /* record to print	     */
-{ pretty_print(rec,"@");			   /*			     */
+static int fput_char(c)				   /*                        */
+  int c;					   /*                        */
+{ return fputc(c,ofile);			   /*                        */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:	print_commented()
-** Purpose:	Print a record to the assigned output file, but print ### 
-**		instead of the eading @.
+** Function:	fput_record()
+** Purpose:	Format and print a complete record onto a given stream.
+**		for further details see |put_record()|.
 ** Arguments:
-**	rec	record to print
+**	file	Stream to print onto.
+**	db	Database containing the record.
+**	rec	Record to print.
+**	start	Initial string used before the type. Should be "@" normally.
 ** Returns:	nothing
 **___________________________________________________			     */
-void print_commented(rec)			   /*			     */
-  register Record rec;				   /* record to print	     */
-{ pretty_print(rec,"###");			   /*			     */
+void fput_record(file,rec,db,start)	   	   /*			     */
+  FILE	 *file;			   		   /*                        */
+  DB	 db;			   		   /*                        */
+  Record rec;			   		   /* record to print	     */
+  char	 *start;		   	   	   /* initial string = "@"   */
+{						   /*                        */
+  ofile = file;					   /*                        */
+  put_record(fput_char,rec,db,start);		   /*                        */
+}						   /*------------------------*/
+
+ static StringBuffer *osb = NULL;
+
+/*-----------------------------------------------------------------------------
+** Function:	sput_char()
+** Purpose:	Output function which places the character on the |osb| 
+**		string buffer.
+** Arguments:
+**	c	Character to print.
+** Returns:	The return status of |fputc()|.
+**___________________________________________________			     */
+static int sput_char(c)				   /*                        */
+  int c;					   /*                        */
+{ return sbputchar(c,osb);			   /*                        */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:	pretty_print()
+** Function:	sput_record()
+** Purpose:	Format and print a complete record into a string and return it.
+**		The string returned points to static memory which is
+**		reused upon the next invocation of this function.
+** Arguments:
+**	file	Stream to print onto.
+**	db	Database containing the record.
+**	rec	Record to print.
+**	start	Initial string used before the type. Should be "@" normally.
+** Returns:	The string containing the printed representation.
+**___________________________________________________			     */
+char * sput_record(rec,db,start)	   	   /*			     */
+  DB	 db;			   		   /*                        */
+  Record rec;			   		   /* record to print	     */
+  char	 *start;		   	   	   /* initial string = "@"   */
+{						   /*                        */
+  if ( osb == NULL ) osb = sbopen();		   /*                        */
+  sbrewind(osb);				   /*                        */
+  put_record(sput_char,rec,db,start);		   /*                        */
+  return sbflush(osb);
+}						   /*------------------------*/
+
+/*-----------------------------------------------------------------------------
+** Function:	put_record()
 ** Purpose:	Format and print a complete record.
-**		The record type and several resources are taken into acc.
+**		The record type and several resources are taken into
+**		account. The following external variables (from
+**		|rsc.c|) are taken into account:
+**		\begin{description}
+**		\item[rsc\_parentheses] If this boolean variable is
+**		  |TRUE| then |(| and |)| are used to delimit the
+**		  record. Otherwise |{| and |}| are used.
+**		\item[rsc\_col\_p] This integer variable controlls the
+**		  indentation of preamble records.
+**		\item[rsc\_col\_s] This integer variable controlls the
+**		  indentation of string records.
+**
+**		\item[rsc\_expand\_macros] If this boolean variable is
+**		  set then macros are expanded before the record is
+**		  printed. This does not effect the internal
+**		  representation. 
+**		\item[rsc\_col] This integer variable controlls the
+**		  indentation of normal records.
+**		\item[rsc\_col\_key] This integer variable controlls the
+**		  indentation of the key in a normal record.
+**		\item[rsc\_newlines] This integer variable controlls
+**		  the number of newlines printed after a normal record.
+**
+**		\item[rsc\_linelen] This integer variable controlls
+**		  the length of the line. The line breaking algorithm
+**		  is applied if this column is about to be violated.
+**		\item[rsc\_indent] This integer variable controlls the
+**		  indentation of equations.
+**		\item[rsc\_eq\_right] This boolean variable controlls
+**		  the alignment of the |=| in equations. It it is set
+**		  then the equality sign is flused right. Otherwise it
+**		  is flushed left.
+**		\end{description}
+**
+**		The field in the record are sorted with
+**		|sort_record()| before they are printed.
+**
+**		In normal records all fields not starting with an
+**		allowed character are ignored. Thus it is possible to
+**		store private and invisible information in a
+**		field. Simply start the field name with an not allowed
+**		character like |%|.
 ** Arguments:
-**	rec	record to print
-**	at	initial string used before the type. Should be "@" normally.
+**	fct	function to use for writing a character.
+**	db	Database containing the record.
+**	rec	Record to print.
+**	start	Initial string used before the type. Should be "@" normally.
 ** Returns:	nothing
 **___________________________________________________			     */
-static void pretty_print(rec,at)		   /*			     */
-  register Record	rec;			   /* record to print	     */
-  char			*at;			   /* initial string = "@"   */
-{ register char		**hp;			   /* heap pointer	     */
-  register unsigned int i;			   /*			     */
+void put_record(fct,rec,db,start)		   /*                        */
+  int    (*fct)_ARG((int));			   /*                        */
+  Record rec;					   /*                        */
+  DB	 db;			   		   /*                        */
+  char	 *start;		   	   	   /* initial string = "@"   */
+{ char	       **hp;			   	   /* heap pointer	     */
+  unsigned int i;			   	   /*			     */
   char open_brace, close_brace;			   /*			     */
- 						   /*                        */
+  						   /*                        */
   sort_record(rec);				   /*                        */
  						   /*                        */
   hp = RecordHeap(rec);				   /*			     */
-  if ( IsNormalRecord(RecordToken(rec)) ) { NL; }  /*			     */
-  PUTS(at);					   /*			     */
-  PUTS(EntryName(RecordToken(rec)));		   /*			     */
+  if ( IsNormalRecord(RecordType(rec)) ) { NL; }   /*			     */
+ 						   /*                        */
+  if ( *RecordComment(rec) )		   	   /*                        */
+  { PUTS(RecordComment(rec));			   /*                        */
+    NL;						   /*                        */
+  }						   /*                        */
  						   /*                        */
   if ( rsc_parentheses )			   /*                        */
   { open_brace  = '(';			   	   /*                        */
@@ -440,49 +513,87 @@ static void pretty_print(rec,at)		   /*			     */
     close_brace = '}';			   	   /*                        */
   }						   /*                        */
  						   /*                        */
-  switch ( RecordToken(rec) )			   /*			     */
-  { case COMMENT:				   /*			     */
-      indent(rsc_col_c);			   /*			     */
+  switch ( RecordType(rec) )			   /*			     */
+  { case BIB_COMMENT:				   /*			     */
+#ifdef OLD
+      indent(rsc_col_c,fct);			   /*			     */
       PUTS(*hp);				   /*			     */
       PUTC(' ');				   /*                        */
       NL;				   	   /*			     */
+#endif
       break;					   /*			     */
-    case PREAMBLE:				   /*			     */
+    case BIB_PREAMBLE:				   /*			     */
+      PUTS(start);				   /*			     */
+      PUTS(EntryName(RecordType(rec)));	   	   /*			     */
       PUTC(open_brace);				   /*			     */
-      indent(rsc_col_p);			   /*			     */
-      line_breaking(*hp,rsc_col_p);		   /*			     */
+      indent(rsc_col_p,fct);			   /*			     */
+      line_breaking(*hp,rsc_col_p,fct);		   /*			     */
       PUTC(' ');				   /*                        */
       PUTC(close_brace);			   /*                        */
       NL;				   	   /*			     */
       break;					   /*			     */
-    case STRING:				   /*			     */
+    case BIB_STRING:				   /*			     */
+      PUTS(start);				   /*			     */
+      PUTS(EntryName(RecordType(rec)));	   	   /*			     */
       PUTC(open_brace);				   /*			     */
-      print_equation(*hp,*(hp+1),rsc_col_s);	   /*			     */
+      print_equation(sym_empty,*hp,*(hp+1),rsc_col_s,fct);/*		     */
       PUTC(' ');				   /*                        */
       PUTC(close_brace);			   /*                        */
       NL;				   	   /*			     */
       break;					   /*			     */
+    case BIB_ALIAS:				   /*			     */
+      PUTS(start);				   /*			     */
+      PUTS(EntryName(RecordType(rec)));	   	   /*			     */
+      PUTC(open_brace);				   /*			     */
+      print_equation(sym_empty,*hp,*(hp+1),rsc_col_s,fct);/*		     */
+      PUTC(' ');				   /*                        */
+      PUTC(close_brace);			   /*                        */
+      NL;				   	   /*			     */
+      break;					   /*			     */
+    case BIB_MODIFY:				   /*			     */
+    case BIB_INCLUDE:				   /*			     */
+      NL;				   	   /*			     */
+      break;					   /*			     */
     default:					   /*			     */
+      PUTS(start);				   /*			     */
+      PUTS(EntryName(RecordType(rec)));	   	   /*			     */
       PUTC(open_brace);				   /*			     */
-      for ( i=RecordFree(rec); i>0; i-=2 )	   /*			     */
-      { if ( *hp )				   /* No deleted entry	     */
-	{ if ( *(hp+1) )			   /* If equation	     */
-	  { at = ( rsc_expand_macros		   /*			     */
-		  ? expand_rhs(*(hp+1),rsc_braces) /*			     */
-		  : *(hp+1) );			   /*			     */
-	    print_equation(*hp,			   /*			     */
-			   at,			   /*			     */
-			   rsc_col);		   /*			     */
+      { char *comma1 = sym_empty,		   /*                        */
+	     *comma2 = ",";			   /*                        */
+ 						   /*                        */
+        if (rsc_print_ce)			   /*                        */
+	{ comma1 = comma2;			   /*                        */
+	  comma2 = sym_empty;			   /*                        */
+	}					   /*                        */
+ 						   /*                        */
+        for ( i=RecordFree(rec); i>0; i-=2 )	   /*			     */
+	{		   			   /* No deleted or          */
+	  if ( *hp && is_allowed(**hp) )	   /*   private entry        */
+	  { 					   /*                        */
+	    if ( *(hp+1) )			   /* If equation	     */
+	    { PUTS(comma1);			   /*                        */
+	      NL;			   	   /*                        */
+	      print_equation(comma2,		   /*                        */
+			     *hp,		   /*                        */
+			     (rsc_expand_macros	   /*                        */
+			      ? expand_rhs(*(hp+1),/*                        */
+					   (rsc_braces?"{":"\""),/*          */
+					   (rsc_braces?"}":"\""),/*          */
+					   db)	   /*                        */
+			      : *(hp+1) ),	   /*                        */
+			     rsc_col,		   /*                        */
+			     fct);		   /*                        */
+	    }					   /*			     */
+	    else				   /* Otherwise print a key  */
+	    { indent(rsc_col_key,fct);		   /*			     */
+	      PUTS(get_key_name(*hp));      	   /*			     */
+	    }					   /*			     */
 	  }					   /*			     */
-	  else					   /* Otherwise print a key  */
-	  { indent(rsc_col_key);		   /*			     */
-	    PUTS(get_key_name(*hp));      	   /*			     */
-	  }					   /*			     */
-	  if ( i > 2 ) { PUTC(','); NL; }	   /* write sep if required  */
-	}					   /*			     */
-	hp+=2;					   /* Goto next pair.	     */
+	  hp += 2;				   /* Goto next pair.	     */
+	}					   /*                        */
       }						   /*			     */
-      NL; PUTC(close_brace);	   	   	   /*                        */
+      NL;					   /*                        */
+      PUTC(close_brace);	   	   	   /*                        */
       for ( i=rsc_newlines; i>0; --i ) { NL; }	   /*                        */
   }						   /*			     */
 }						   /*------------------------*/
