@@ -1,27 +1,42 @@
 /******************************************************************************
-** $Id: record.c,v 2.14 1996/02/22 21:43:51 gerd Exp gerd $
+** $Id: record.c,v 2.21 1997/12/03 11:50:30 gerd Exp $
 **=============================================================================
 ** 
 ** This file is part of BibTool.
 ** It is distributed under the GNU General Public License.
 ** See the file COPYING for details.
 ** 
-** (c) 1996 Gerd Neugebauer
+** (c) 1996-1997 Gerd Neugebauer
 ** 
 ** Net: gerd@informatik.uni-koblenz.de
 ** 
 ******************************************************************************/
 
-#include "bibtool.h"
-#include "error.h"
-#include "type.h"
-#include "record.h"
-#include "symbols.h"
+#include <bibtool/general.h>
+#include <bibtool/error.h>
+#include <bibtool/record.h>
+#include <bibtool/symbols.h>
 
 /*****************************************************************************/
 /* Internal Programs							     */
 /*===========================================================================*/
 
+#ifdef __STDC__
+#define _ARG(A) A
+#else
+#define _ARG(A) ()
+#endif
+ Record copy_record _ARG((Record rec));		   /* record.c               */
+ Record new_record _ARG((int token,int size));	   /* record.c               */
+ Record record_gc _ARG((Record rec));		   /* record.c               */
+ Record unlink_record _ARG((Record rec));	   /* record.c               */
+ WordList new_wordlist _ARG((char * s));	   /* record.c               */
+ void add_sort_order _ARG((char *val));		   /* record.c               */
+ void free_1_record _ARG((Record rec));		   /* record.c               */
+ void free_record _ARG((Record rec));		   /* record.c               */
+ void push_to_record _ARG((Record rec,char *s,char *t));/* record.c          */
+ void sort_record _ARG((Record rec));		   /* record.c               */
+
 /*****************************************************************************/
 /* External Programs							     */
 /*===========================================================================*/
@@ -30,10 +45,12 @@
 
 /*-----------------------------------------------------------------------------
 ** Function:	copy_record()
-** Purpose:	Copy a record.
+** Purpose:	Copy a record and return a new instance.
+**		If no memory is left then an error is raised and the
+**		program is terminated. 
 ** Arguments:
-**	rec
-** Returns:	
+**	rec	The record to copy.
+** Returns:	The new copy of |rec|.
 **___________________________________________________			     */
 Record copy_record(rec)				   /*			     */
   register Record rec;				   /*			     */
@@ -46,13 +63,16 @@ Record copy_record(rec)				   /*			     */
      || (new_heap=(char**)malloc(sizeof(char*)*(size_t)RecordFree(rec)))/*   */
 	== 0L )					   /*			     */
   { OUT_OF_MEMORY("Record"); }      		   /*                        */
-  RecordKey(new)	  = "";			   /*			     */
+  RecordSortkey(new)	  = sym_empty;		   /*			     */
   RecordOldKey(new)	  = RecordOldKey(rec);	   /*			     */
-  NextRecord(new)	  = RecNULL;		   /*			     */
-  PrevRecord(new)	  = RecNULL;		   /*			     */
-  Record_Full_Token(new)  = Record_Full_Token(rec);/*			     */
+  NextRecord(new)	  = RecordNULL;		   /*			     */
+  PrevRecord(new)	  = RecordNULL;		   /*			     */
+  RecordType(new)	  = RecordType(rec);	   /*			     */
   RecordFree(new)	  = RecordFree(rec);	   /*			     */
+  RecordComment(new)	  = RecordComment(rec);	   /*			     */
+  RecordSource(new)	  = RecordSource(rec);	   /*			     */
   RecordHeap(new)	  = new_heap;		   /*			     */
+  RecordFlags(new)	  = RecordFlags(rec);	   /*			     */
   for ( i=0, old_heap=RecordHeap(rec);		   /*			     */
 	i < RecordFree(new);			   /*			     */
 	++i )					   /*			     */
@@ -60,87 +80,217 @@ Record copy_record(rec)				   /*			     */
   return (new);					   /*			     */
 }						   /*------------------------*/
 
+/*-----------------------------------------------------------------------------
+** Function:	new_record()
+** Purpose:	Create a new record and return it.
+**		If no memory is left then an eror is raised and the
+**		program is terminated. 
+** Arguments:
+**	token	The token type of the record.
+**	size	The initial heap size.
+** Returns:	The new record.
+**___________________________________________________			     */
+Record new_record(token,size)			   /*			     */
+  int token;				   	   /*			     */
+  int size;					   /*                        */
+{ register Record new;				   /*			     */
+  register char	 **new_heap;			   /*			     */
+  register int	 i;				   /*			     */
+						   /*			     */
+  if ( size < 1 ) size = 1;			   /*                        */
+  if (	(new=(Record)malloc(sizeof(SRecord))) == 0L/*			     */
+     || (new_heap=(char**)malloc(sizeof(char*)*(size_t)(size)))/*            */
+	== 0L )					   /*			     */
+  { OUT_OF_MEMORY("Record"); }      		   /*                        */
+  RecordSortkey(new)	  = sym_empty;		   /*			     */
+  RecordOldKey(new)	  = RecordSortkey(new);	   /*			     */
+  NextRecord(new)	  = RecordNULL;		   /*			     */
+  PrevRecord(new)	  = RecordNULL;		   /*			     */
+  RecordType(new)	  = token;		   /*			     */
+  RecordFree(new)	  = size;		   /*			     */
+  RecordComment(new)	  = sym_empty;		   /*			     */
+  RecordSource(new)	  = sym_empty;		   /*			     */
+  RecordFlags(new)	  = 0;	   		   /*			     */
+  RecordHeap(new)	  = new_heap;		   /*			     */
+  for ( i=0; i < RecordFree(new); ++i )		   /*			     */
+  { *(new_heap++) = NULL; }		   	   /*			     */
+  return (new);					   /*			     */
+}						   /*------------------------*/
+
+/*-----------------------------------------------------------------------------
+** Function:	free_record()
+** Purpose:	Release a list of records. All records reachable through a
+**		previous/next chain are deallocated.
+** Arguments:
+**	rec	Arbitrary Record in the chain.
+** Returns:	nothing
+**___________________________________________________			     */
+void free_record(rec)				   /*                        */
+  Record rec;					   /*                        */
+{ Record r;					   /*                        */
+ 						   /*                        */
+  if ( rec == RecordNULL ) return;		   /*                        */
+ 						   /*                        */
+  while ( PrevRecord(rec) != RecordNULL )	   /* rewind                 */
+  { rec = PrevRecord(rec); }			   /*                        */
+ 						   /*                        */
+  while ( rec != RecordNULL )			   /*                        */
+  {						   /*                        */
+    r = rec;					   /*                        */
+    rec = NextRecord(rec);			   /*                        */
+    free_1_record(r);				   /*                        */
+  }						   /*                        */
+}						   /*------------------------*/
+
+/*-----------------------------------------------------------------------------
+** Function:	free_1_record()
+** Purpose:	Free the memory occupied by a single record.
+**		This does not ensure that there is no dangling pointer to
+**		the record. Thus beware!
+** Arguments:
+**	rec	record to free
+** Returns:	nothing
+**___________________________________________________			     */
+void free_1_record(rec)				   /*                        */
+  Record rec;					   /*                        */
+{ int i;					   /*                        */
+ 						   /*                        */
+  if ( rec != RecordNULL )			   /*                        */
+  {						   /*                        */
+    ReleaseSymbol(RecordSortkey(rec));		   /*                        */
+    ReleaseSymbol(RecordOldKey(rec));		   /*                        */
+    ReleaseSymbol(RecordComment(rec));		   /*                        */
+    ReleaseSymbol(RecordSource(rec));		   /*                        */
+    if ( RecordHeap(rec) != NULL )		   /*                        */
+    {						   /*                        */
+      for ( i=0; i<RecordFree(rec); i++ )	   /*                        */
+      { ReleaseSymbol(RecordHeap(rec)[i]); }	   /*                        */
+      free(RecordHeap(rec));			   /*                        */
+    }						   /*                        */
+    free(rec);					   /*                        */
+  }						   /*                        */
+}						   /*------------------------*/
+
+/*-----------------------------------------------------------------------------
+** Function:	unlink_record()
+** Purpose:	Remove a record from a chain and free its memory. The
+**		chain is modified such that the freed Record is not
+**		referenced any more. A neighbor in the chain of the
+**		given record is returned or |NULL| if there is none.
+** Arguments:
+**	rec	Record to free.
+** Returns:	nothing
+**___________________________________________________			     */
+Record unlink_record(rec)			   /*                        */
+  Record rec;					   /*                        */
+{ Record r;					   /*                        */
+ 						   /*                        */
+  if ( NextRecord(rec) )			   /*                        */
+  { PrevRecord(NextRecord(rec)) = PrevRecord(rec); /*                        */
+    if (PrevRecord(rec))			   /*                        */
+    { NextRecord(PrevRecord(rec)) = NextRecord(rec);/*                       */
+    }						   /*                        */
+    r = NextRecord(rec);			   /*                        */
+  }						   /*                        */
+  else if (PrevRecord(rec))			   /*                        */
+  { NextRecord(PrevRecord(rec)) = NextRecord(rec); /*                        */
+    r = PrevRecord(rec);			   /*                        */
+  }						   /*                        */
+  else						   /*                        */
+  { r = RecordNULL;				   /*                        */
+  }						   /*                        */
+  free_1_record(rec);				   /*                        */
+  return r;					   /*                        */
+}						   /*------------------------*/
+
+/*-----------------------------------------------------------------------------
+** Function:	record_gc()
+** Purpose:	Garbage collecting a record list. The entries marked
+**		as deleted are unlinked and the memory is freed. Any
+**		pointers to such a deleted entry become invalid.
+**		
+**		Be careful when using this function!
+** Arguments:
+**	rec	Pointer to any entry in the chain.
+** Returns:	Pointer to some entry in the cleared chain or
+**		|RecordNULL| if none is left.
+**___________________________________________________			     */
+Record record_gc(rec)				   /*                        */
+  Record rec;					   /*                        */
+{ Record ret = RecordNULL;			   /*                        */
+ 						   /*                        */
+  if ( rec == RecordNULL ) return rec;
+  while ( PrevRecord(rec) != RecordNULL )
+  { rec = PrevRecord(rec); }
+ 
+  while ( rec != RecordNULL )
+  {
+    if ( RecordIsDELETED(rec) )
+    { rec = unlink_record(rec);
+    }
+    else
+    { ret = rec;
+      rec = NextRecord(rec);
+    }
+  }
+
+  return ret;					   /*                        */
+}						   /*------------------------*/
+
+/*-----------------------------------------------------------------------------
+** Function:	push_to_record()
+** Purpose:	Put an equation s=t onto the heap of a record.
+**		If a field s is already there then the value is
+**		overwritten.  The arguments are expected to be
+**		synbols. Thus it is not necessary to make private
+**		copies and it is possible to avoid expensive string
+**		comparisons. 
+** Arguments:
+**	s	Left hand side of the equation.
+**	t	Right hand side of the equation.
+** Returns:	nothing
+**___________________________________________________			     */
+void push_to_record(rec,s,t)			   /*			     */
+  register Record rec;				   /*                        */
+  register char *s;				   /*			     */
+  register char *t;				   /*			     */
+{ register int i;		   		   /*			     */
+   						   /*                        */
+  if ( s == sym_crossref ) { SetRecordXREF(rec); } /*			     */
+ 						   /*                        */
+  for ( i=2; i < RecordFree(rec); i+=2 )	   /* search the field       */
+  { if ( RecordHeap(rec)[i] == s )		   /* if found then          */
+    { RecordHeap(rec)[i+1] = t;			   /* overwrite the value    */
+      return;					   /*                        */
+    }   					   /*                        */
+  }						   /*                        */
+  for ( i=2; i < RecordFree(rec); i+=2 )	   /* search empty field     */
+  { if ( RecordHeap(rec)[i] == (char*)0 )	   /* if found then          */
+    { RecordHeap(rec)[i++] = s;			   /* add the new item       */
+      RecordHeap(rec)[i]   = t;			   /*                        */
+      return;					   /*                        */
+    }   					   /*                        */
+  }						   /*                        */
+  i = RecordFree(rec);				   /*                        */
+  RecordFree(rec) += 2;				   /*                        */
+  if ( (RecordHeap(rec)   			   /* enlarge the heap	     */
+	=(char**)realloc(RecordHeap(rec),	   /*	                     */
+			 RecordFree(rec)*sizeof(char*)))/*                   */
+     == (char**)0 )				   /*	                     */
+  { OUT_OF_MEMORY("heap"); }      		   /*                        */
+						   /*			     */
+  RecordHeap(rec)[i++] = s;		   	   /*			     */
+  RecordHeap(rec)[i]   = t;		   	   /*			     */
+}						   /*------------------------*/
+
 
 /*****************************************************************************/
 /***			                         			   ***/
 /*****************************************************************************/
 
-#define HEAP_INIT 16
-#define HEAP_INC  8
-
- static Record	master_record = RecordNULL;
- static size_t	master_size   = 0;
- char		*sym_crossref = (char*)0;
-
-/*-----------------------------------------------------------------------------
-** Function:	clear_master_record()
-** Purpose:	This function initializes this master record properly.
-**		The master record is a static Record structure used by
-**		the parser to deposit its intermediate results.
-** Arguments:
-**	token
-** Returns:	nothing
-**___________________________________________________			     */
-void clear_master_record(token)			   /*			     */
-  register int token;				   /*			     */
-{						   /*			     */
-  if ( master_record == RecordNULL )		   /*			     */
-  { master_size = HEAP_INIT;			   /*			     */
-    if (   RecordNULL == (master_record		   /*			     */
-			  =(Record)malloc(sizeof(SRecord)))/*		     */
-	|| (char**)0  == (RecordHeap(master_record)/*			     */
-			  =(char**)malloc(master_size*sizeof(char*))) )/*    */
-    { OUT_OF_MEMORY("master record"); }		   /*                        */
-    sym_crossref = sym_add("crossref",-1);	   /*			     */
-  }						   /*			     */
-  RecordOldKey(master_record)	   = NULL;	   /*			     */
-  Record_Full_Token(master_record) = token;	   /*			     */
-  RecordFree(master_record)	   = 0;		   /*			     */
-}						   /*------------------------*/
-
-/*-----------------------------------------------------------------------------
-** Function:	push_to_master_record()
-** Purpose:	Put an equation s=t onto the master record.
-** Arguments:
-**	s
-**	t
-** Returns:	nothing
-**___________________________________________________			     */
-void push_to_master_record(s,t)			   /*			     */
-  register char *s;				   /*			     */
-  register char *t;				   /*			     */
-{ register int i;				   /*			     */
-						   /*			     */
-  if ( (i=RecordFree(master_record)) >= master_size-2 )/*		     */
-  { master_size += HEAP_INC;			   /*			     */
-    if ( (char**)0 == (RecordHeap(master_record)   /*			     */
-		       =(char**)realloc(RecordHeap(master_record),/*	     */
-					master_size*sizeof(char*))) )/*	     */
-    { OUT_OF_MEMORY("heap"); }      		   /*                        */
-  }						   /*			     */
-						   /*			     */
-  RecordHeap(master_record)[i++] = s;		   /*			     */
-  RecordHeap(master_record)[i]	 = t;		   /*			     */
-  RecordFree(master_record)	+= 2;		   /*			     */
-  if ( s == sym_crossref )			   /*			     */
-  { SetRecordXREF(master_record); }		   /*			     */
-}						   /*------------------------*/
-
-/*-----------------------------------------------------------------------------
-** Function:	get_master_record()
-** Purpose:	Return the address of the master record structure.
-** Arguments:	none
-** Returns:	the address
-**___________________________________________________			     */
-Record get_master_record()			   /*			     */
-{ return master_record;				   /*			     */
-}						   /*------------------------*/
-
-
-
-#include "wordlist.h"
-#include "s_parse.h"
-#include "entry.h"
+#include <bibtool/wordlist.h>
+#include <bibtool/s_parse.h>
+#include <bibtool/entry.h>
  
 
 /*-----------------------------------------------------------------------------
@@ -190,15 +340,15 @@ void add_sort_order(val)			   /*                        */
   OrderList ol;					   /*                        */
   WordList  *wlp, wl, wl_next;			   /*                        */
  						   /*                        */
-  SParseSkip(&val);				   /*                        */
+  (void)SParseSkip(&val);			   /*                        */
   if ( *val == '*' )				   /*                        */
   { s    = NULL;				   /*                        */
-    type = NOOP;				   /*                        */
+    type = BIB_NOOP;				   /*                        */
     val++;					   /*                        */
   }						   /*                        */
   else if ( (s=SParseSymbol(&val)) == NULL )	   /*                        */
   { return; }					   /*                        */
-  else if ( (type=find_entry_type(s)) == NOOP )	   /*                        */
+  else if ( (type=find_entry_type(s)) == BIB_NOOP )/*                        */
   { Err("Undefined entry type for sort order");	   /*                        */
     return;					   /*                        */
   }						   /*                        */
@@ -220,7 +370,7 @@ void add_sort_order(val)			   /*                        */
     wlp           = &OrderVal(ol);		   /*                        */
   }						   /*                        */
    						   /*                        */
-  SParseSkip(&val);				   /*                        */
+  (void)SParseSkip(&val);			   /*                        */
   						   /*                        */
   while ( *val )				   /*                        */
   { if ( (s=SParseSymbol(&val)) != NULL )	   /*                        */
@@ -260,7 +410,7 @@ void sort_record(rec)				   /*			     */
 { Record          r;				   /* Temp. record           */
   OrderList       ol;				   /*                        */
   int             i, ptr;			   /*                        */
-  int             type = RecordToken(rec);	   /*                        */
+  int             type = RecordType(rec);	   /*                        */
   WordList        wl;				   /*                        */
  						   /*                        */
   for ( ol=order;				   /* Search for an order    */
@@ -268,7 +418,7 @@ void sort_record(rec)				   /*			     */
        ol = NextOrder(ol) ) {}			   /*  record first.         */
   if ( ol == OrderNULL )			   /* If none was found then */
   { for ( ol=order;				   /* Search for an order    */
-	 ol && OrderType(ol) != NOOP;		   /*  applicable for all    */
+	 ol && OrderType(ol) != BIB_NOOP;	   /*  applicable for all    */
 	 ol = NextOrder(ol) ) {}		   /*  record types.         */
     if ( ol == OrderNULL ) return;		   /* No order then return.  */
   }						   /*                        */
