@@ -1,12 +1,12 @@
 /******************************************************************************
-** $Id: tex_read.c,v 2.17 1996/02/22 21:43:51 gerd Exp gerd $
+** $Id: tex_read.c,v 2.20 1997/12/03 11:50:30 gerd Exp $
 **=============================================================================
 ** 
 ** This file is part of BibTool.
 ** It is distributed under the GNU General Public License.
 ** See the file COPYING for details.
 ** 
-** (c) 1996 Gerd Neugebauer
+** (c) 1996-1997 Gerd Neugebauer
 ** 
 ** Net: gerd@informatik.uni-koblenz.de
 **	    
@@ -14,25 +14,30 @@
 ** Purpose: Simulate the TeX reading mechanism.
 **
 ** Compilation: cc tex_read.c -o tex_read -DSTANDALONE
+** 
+**-----------------------------------------------------------------------------
+** Description:
+**	This module contains functions which immitate the reading
+**	apparatus of \TeX{}. Macro expansion can be performed. 
+** 
 **	    
 ******************************************************************************/
 
-#include "config.h"
-#include <stdio.h>
+#include <bibtool/general.h>
 #include <ctype.h>
 
+#ifndef STANDALONE
+#include <bibtool/error.h>
+#else
+#define Err(X)		 (void)fputs(X,stderr)
+#define ERROR_EXIT(X)	 Err(X); exit(1)
+#define OUT_OF_MEMORY(X) Err("Out of memory for "); Err(X); exit(1)
+#endif
+
 /*****************************************************************************/
 /* External Programs							     */
 /*===========================================================================*/
 
-#ifdef __STDC__
-#include <stdlib.h>
-#include <string.h>
-#else
- extern VoidPTR malloc();
-#endif
-
-
 #define CARET_CARET	    /* enable ^^A type specification of unprintables */
 #undef	TRANSLATE_NEWLINE   /* Translate nl to space or \par */
 
@@ -74,7 +79,7 @@
  static MacDef find_macro _ARG((char *name,MacDef md));/* tex_read.c         */
  static MacDef new_macdef _ARG((char *name,int arity,Token tokens));/* tex_read.c*/
  static Token TeX_get_token _ARG((int (*get_fct)_ARG((void))));/* tex_read.c */
- static Token new_token _ARG((int type,char *string)); /* tex_read.c         */
+ static Token new_token _ARG((int type,char *string));/* tex_read.c          */
  static Token token_list_copy _ARG((Token t,Token nt,Token *argp));/* tex_read.c*/
  static Token tokenize _ARG((char *s,int arity));  /* tex_read.c             */
  static char * tokens_to_string _ARG((Token t));   /* tex_read.c             */
@@ -85,6 +90,7 @@
  static int get_file _ARG((void));		   /* tex_read.c             */
  static int get_string _ARG((void));		   /* tex_read.c             */
  static void free_1_token _ARG((Token t));	   /* tex_read.c             */
+ static void free_macdef _ARG((MacDef mac));	   /* tex_read.c             */
  static void free_tokens _ARG((Token t));	   /* tex_read.c             */
  static void init_TeX _ARG((void));		   /* tex_read.c             */
  static void init_get _ARG((char * s));		   /* tex_read.c             */
@@ -94,6 +100,7 @@
  void TeX_define _ARG((char *name,int arity,char *body));/* tex_read.c       */
  void TeX_open_file _ARG((FILE * file));	   /* tex_read.c             */
  void TeX_open_string _ARG((char * s));		   /* tex_read.c             */
+ void TeX_reset _ARG((void));			   /* tex_read.c             */
 
 #ifdef STANDALONE
  int main _ARG((int argc,char *argv[]));	   /* tex_read.c             */
@@ -105,17 +112,6 @@
 
 /*---------------------------------------------------------------------------*/
 
-#define TRUE  (0)
-#define FALSE (-1)
-
-#ifdef STANDALONE
-#define Err(X) (void)fputs(X,stderr)
-#define ERROR_EXIT(X) Err(X);exit(1)
-#define OUT_OF_MEMORY(X) Err("Out of memory for ");Err(X);exit(1)
-#else
-#include "error.h"
-#endif
-
 /*****************************************************************************/
 /*** catcode Management.						   ***/
 /*****************************************************************************/
@@ -150,7 +146,7 @@
 #define CATCODE_COMMENT		0x0E00
 #define CATCODE_INVALID		0x0F00
 
-#define CATCODE_MASK 0x0F00
+#define CATCODE_MASK		0x0F00
 
  static short int catcode[256];
 
@@ -161,17 +157,21 @@
 
 /*-----------------------------------------------------------------------------
 ** Function:	init_TeX()
-** Purpose:	Initialize the TeX reading apparatus.
-**		Mainly the catcodes are assigned.
+** Purpose:	Initialize the \TeX{} reading apparatus.
+**		Mainly the catcodes are assigned and the macros are
+**		cleared.  This function has to be called before the
+**		other functions in this module. Thus this function  is
+**		called automatically from different places to ensure
+**		that the reading apparatus is initialized.
 ** Arguments:	none
 ** Returns:	nothing
 **___________________________________________________			     */
 static void init_TeX()				   /*			     */
 { register int i;				   /*			     */
-  static int   first = TRUE;			   /*			     */
+  static int   initialized = FALSE;		   /*			     */
 						   /*			     */
-  if ( !first ) return;				   /*			     */
-  first = FALSE;				   /*			     */
+  if ( initialized ) return;			   /*			     */
+  initialized = TRUE;				   /*			     */
 						   /*			     */
   for ( i=0; i<256; ++i )			   /*			     */
   { active[i]  = MacDefNULL;			   /*			     */
@@ -179,19 +179,19 @@ static void init_TeX()				   /*			     */
 		  ? CATCODE_LETTER		   /*			     */
 		  : CATCODE_OTHER  );		   /*			     */
   }						   /*			     */
-  catcode['\\'] = CATCODE_ESCAPE;		   /*			     */
-  catcode['{']	= CATCODE_BEG_GROUP;		   /*			     */
-  catcode['}']	= CATCODE_END_GROUP;		   /*			     */
-  catcode['$']	= CATCODE_MATH;			   /*			     */
-  catcode['&']	= CATCODE_ALIGN;		   /*			     */
-  catcode['\n'] = CATCODE_EOL;			   /*			     */
-  catcode['#']	= CATCODE_PARAMETER;		   /*			     */
-  catcode['^']	= CATCODE_SUPERSCRIPT;		   /*			     */
-  catcode['_']	= CATCODE_SUBSCRIPT;		   /*			     */
-  catcode['\0'] = CATCODE_IGNORED;		   /*			     */
-  catcode[' ']	= CATCODE_SPACE;		   /*			     */
-  catcode['~']	= CATCODE_ACTIVE;		   /*			     */
-  catcode['%']	= CATCODE_COMMENT;		   /*			     */
+  catcode[CHAR_ESCAPE]	    = CATCODE_ESCAPE;	   /*			     */
+  catcode[CHAR_BEG_GROUP]   = CATCODE_BEG_GROUP;   /*			     */
+  catcode[CHAR_END_GROUP]   = CATCODE_END_GROUP;   /*			     */
+  catcode[CHAR_MATH]	    = CATCODE_MATH;	   /*			     */
+  catcode[CHAR_ALIGN]	    = CATCODE_ALIGN;	   /*			     */
+  catcode[CHAR_EOL]	    = CATCODE_EOL;	   /*			     */
+  catcode[CHAR_PARAMETER]   = CATCODE_PARAMETER;   /*			     */
+  catcode[CHAR_SUPERSCRIPT] = CATCODE_SUPERSCRIPT; /*			     */
+  catcode[CHAR_SUBSCRIPT]   = CATCODE_SUBSCRIPT;   /*			     */
+  catcode[CHAR_IGNORED]     = CATCODE_IGNORED;	   /*			     */
+  catcode[CHAR_SPACE]	    = CATCODE_SPACE;	   /*			     */
+  catcode[CHAR_COMMENT]	    = CATCODE_COMMENT;	   /*			     */
+  catcode['~']		    = CATCODE_ACTIVE;	   /*			     */
 }						   /*------------------------*/
  
 /*****************************************************************************/
@@ -206,7 +206,7 @@ static void init_TeX()				   /*			     */
 ** Arguments:
 **	type
 **	string
-** Returns:	
+** Returns:	the new token
 **___________________________________________________			     */
 static Token new_token(type,string)		   /*			     */
   int	   type;				   /*			     */
@@ -481,9 +481,6 @@ static Token TeX_get_token(get_fct)		   /*			     */
   }						   /*			     */
 }						   /*------------------------*/
 
-
-
-
  static char *g_s;
  static char *g_p;
 
@@ -539,7 +536,7 @@ static Token tokenize(s,arity)			   /*			     */
   { if ( TokenChar(nt) =='\n' )			   /*			     */
     { free_1_token(nt); }			   /*			     */
     else					   /*			     */
-    { if ( TokenChar(nt) == '#'			   /*			     */
+    { if ( TokenChar(nt) == CHAR_PARAMETER	   /*			     */
 	  && (t0=TeX_get_token(do_get)) != TokenNULL )/*		     */
       { a = TokenChar(t0) - '0';		   /*			     */
 	if ( 0<a && a<=arity ) TokenChar(nt) |= a<<8;/*			     */
@@ -580,6 +577,14 @@ static MacDef new_macdef(name,arity,tokens)	   /*			     */
   return new;					   /*			     */
 }						   /*------------------------*/
 
+static void free_macdef(mac)
+  MacDef mac;
+{
+  if ( mac == MacDefNULL ) return;
+  free(MacroToken(mac));
+  free(mac);
+}
+
 /*-----------------------------------------------------------------------------
 ** Function:	find_macro()
 ** Purpose:	
@@ -603,7 +608,7 @@ static MacDef find_macro(name,md)		   /*			     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	TeX_define()
-** Purpose:	
+** Purpose:	Add a new \TeX{} macro definition.
 **		
 **
 ** Arguments:
@@ -620,7 +625,9 @@ void TeX_define(name,arity,body)		   /*			     */
 						   /*			     */
   if ( 0 > arity || arity > 9 ) return;		   /*			     */
 						   /*			     */
-  md		= new_macdef(name,arity,tokenize(body,arity));/*	     */
+  md		= new_macdef(name,		   /*                        */
+			     arity,		   /*                        */
+			     tokenize(body,arity));/*	                     */
   NextMacro(md) = macro;			   /*			     */
   macro		= md;				   /*			     */
 }						   /*------------------------*/
@@ -629,9 +636,11 @@ void TeX_define(name,arity,body)		   /*			     */
 ** Function:	TeX_def()
 ** Purpose:	Define a macro.
 **		The argument is a string specification of the following form:
-**		    \name[arity]=replacement text
-**		    \name=replacement text
-**		0 <= arity <= 9
+**
+**		    |\name[arity]=replacement text|\\
+**		    |\name=replacement text|
+**
+**		\(0 <= arity <= 9\)
 ** Arguments:
 **	s
 ** Returns:	nothing
@@ -660,21 +669,20 @@ void TeX_def(s)					   /*			     */
 	       arity,				   /*			     */
 	       new_string(s+1));		   /*			     */
     *ep = c;					   /*			     */
-  }
-  else
-  { TeX_active(*name,arity,new_string(s+1));
-  }
+  }						   /*                        */
+  else						   /*                        */
+  { TeX_active(*name,arity,new_string(s+1));	   /*                        */
+  }						   /*                        */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
 ** Function:	TeX_active()
-** Purpose:	
-**		
-**
+** Purpose:	Assign a macro to an active character. If the
+**		character is not active then the catcode is changed.  
 ** Arguments:
-**	c
-**	arity
-**	s
+**	c	Character to make active.
+**	arity	Arity of the macro assigned to the active character.
+**	s	Body of the definition as string.
 ** Returns:	nothing
 **___________________________________________________			     */
 void TeX_active(c,arity,s)			   /*			     */
@@ -683,10 +691,37 @@ void TeX_active(c,arity,s)			   /*			     */
   char * s;					   /*			     */
 {						   /*			     */
   EnsureInit;					   /*			     */
-  active[c] = new_macdef((char*)NULL,arity,tokenize(s,arity));/*	     */
+  if ( active[c] ) free_macdef(active[c]);
+ 						   /*                        */
+  active[c] = new_macdef((char*)NULL,		   /*                        */
+			 arity,			   /*                        */
+			 tokenize(s,arity));	   /*	                     */
 }						   /*------------------------*/
 
-
+/*-----------------------------------------------------------------------------
+** Function:	TeX_reset()
+** Purpose:	Reset the \TeX{} reading apparatus to its initial
+**		state.  All macros and active characters are cleared
+**		and the memory is released. Thus this function can
+**		also be used for this purpose.
+** Arguments:	none
+** Returns:	nothing
+**___________________________________________________			     */
+void TeX_reset()				   /*                        */
+{ int    i;					   /*                        */
+  MacDef md, next;				   /*                        */
+ 						   /*                        */
+  for ( i=0;i<256;i++ )				   /*                        */
+  { free_macdef(active[i]); }			   /*                        */
+ 						   /*                        */
+  md = macro;					   /*                        */
+  macro = MacDefNULL;				   /*                        */
+  while ( md )					   /*                        */
+  { next = NextMacro(md);			   /*                        */
+    free_macdef(md);				   /*                        */
+    md = next;					   /*                        */
+  }						   /*                        */
+}						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
 ** Function:	get_EOF()
@@ -734,7 +769,7 @@ static int get_file()				   /*			     */
 ** Function:	TeX_open_file()
 ** Purpose:	Prepare things to parse from a file.
 ** Arguments:
-**	file
+**	file	File pointer of the file to read from.
 ** Returns:	nothing
 **___________________________________________________			     */
 void TeX_open_file(file)			   /*			     */
@@ -748,7 +783,7 @@ void TeX_open_file(file)			   /*			     */
 ** Function:	TeX_open_string()
 ** Purpose:	Prepare things to parse from a string.
 ** Arguments:
-**	s
+**	s	String to read from.
 ** Returns:	nothing
 **___________________________________________________			     */
 void TeX_open_string(s)				   /*			     */
@@ -760,9 +795,10 @@ void TeX_open_string(s)				   /*			     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	TeX_close()
-** Purpose:	
-** Arguments:
-**	
+** Purpose:	Gracefully terminate the reading of \TeX{} tokens. Any
+**		remaining pieces of text which have already been
+**		consumed are discarted.
+** Arguments:	none
 ** Returns:	nothing
 **___________________________________________________			     */
 void TeX_close()				   /*			     */
@@ -783,8 +819,8 @@ static int fill_token(tp)			   /*			     */
   register Token*tp;				   /*			     */
 { if (	 *tp == TokenNULL			   /*			     */
       && (*tp=TeX_get_token(src_get)) == TokenNULL )/*			     */
-  { return TRUE; }				   /*			     */
-  return FALSE;					   /*			     */
+  { return 0; }				   /*			     */
+  return 1;					   /*			     */
 }						   /*------------------------*/
 
 #define UnlinkToken(T,T2)		T2 = T; T = NextToken(T)
@@ -792,11 +828,13 @@ static int fill_token(tp)			   /*			     */
 
 /*-----------------------------------------------------------------------------
 ** Function:	TeX_read()
-** Purpose:	
+** Purpose:	Read a single Token and return it as a pair consisting
+**		of an ASCII code and possibly a string in case of a
+**		macro token.
 ** Arguments:
-**	cp
-**	sp
-** Returns:	
+**	cp	Pointer to position where the character is stored.
+**	sp	Pointer to position where the string is stored.
+** Returns:	|FALSE| iff everything went right.
 **___________________________________________________			     */
 int TeX_read(cp,sp)				   /*			     */
   char		* cp;				   /*			     */
@@ -826,7 +864,7 @@ int TeX_read(cp,sp)				   /*			     */
     { UnlinkAndFreeToken(t,t2);			   /* delete macro token     */
 						   /*			     */
       while ( fill_token(&t)			   /* While there are more   */
-	     && TokenChar(t) == ' ')		   /*  tokens and == ' '     */
+	     && TokenChar(t) == CHAR_SPACE)	   /*  tokens and == ' '     */
       { UnlinkAndFreeToken(t,t2); }		   /* unlink space	     */
 						   /*			     */
       for ( i=1; i<=MacroArity(mac); ++i )	   /* Fill the argument	     */
@@ -839,12 +877,13 @@ int TeX_read(cp,sp)				   /*			     */
 	{ tp = t;				   /*			     */
 	  d  = 0;				   /*			     */
 	  while ( fill_token(&NextToken(tp))  &&   /* While there are more   */
-		 (TokenChar(NextToken(tp)) != '}'  /*  tokens and no	     */
-		 || d != 0 ) )			   /*  matching '}'	     */
+		 (TokenChar(NextToken(tp))	   /*  tokens                */
+		  != CHAR_END_GROUP  		   /*  and                   */
+		 || d != 0 ) )			   /*  no matching '}'	     */
 	  { tp = NextToken(tp);			   /*  walk ahead and	     */
 	    switch ( TokenChar(tp) )		   /*  count braces	     */
 	    { case CHAR_BEG_GROUP: ++d; break;	   /*			     */
-	      case '}': d--; break;		   /*			     */
+	      case CHAR_END_GROUP: d--; break;	   /*			     */
 	    }					   /*			     */
 	  }					   /*			     */
 	  t2 = NextToken(NextToken(tp));	   /*			     */
@@ -879,11 +918,10 @@ int TeX_read(cp,sp)				   /*			     */
 /*****************************************************************************/
 /***			 Stand alone version     			   ***/
 /*****************************************************************************/
-
 #ifdef STANDALONE
 
 /*-----------------------------------------------------------------------------
-** Function:	new_string()
+** Function*:	new_string()
 ** Purpose:	
 **		
 **
@@ -900,7 +938,7 @@ static char * new_string(s)			   /*			     */
 }						   /*------------------------*/
 
 /*-----------------------------------------------------------------------------
-** Function:    main()
+** Function*:   main()
 ** Purpose:     Main rountine for the stand alone version.
 ** Arguments:
 **	argc	Number of arguments (+1)
@@ -934,7 +972,7 @@ int main(argc,argv)				   /*			     */
   { if ( (file = fopen(argv[1],"r")) == NULL )	   /*			     */
     { ERROR_EXIT("File open error"); }		   /*			     */
     while ( fgets(buffer,1024,file) )		   /*			     */
-    { for(s=buffer;*s;++s) if(*s=='\n') *s='\0';   /*			     */
+    { for(s=buffer;*s;++s) if(*s=='\n') *s = '\0'; /*			     */
       for(s=buffer;*s&&isspace(*s);++s) ;	   /*			     */
       if ( *s == '\\' ) TeX_def(s);		   /*			     */
     }						   /*			     */
